# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import collections
import concurrent
import io
import random
import re
import threading
import time
from abc import ABC, abstractmethod
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime
from types import TracebackType
from typing import (
    Tuple, List, Dict, Callable, Any, Optional, Union, OrderedDict, TypeVar, Type, Mapping,
    Sequence, Generic, Generator
)

import fabric  # type: ignore[import-untyped]
import fabric.transfer  # type: ignore[import-untyped]

import invoke

from kubemarine.core import log, static, errors, connections
from kubemarine.core.connections import ConnectionPool
from kubemarine.core.environment import Environment

EXIT_CODE_PATTERN = re.compile(r'^\n(\d+)\n')


class RunnersResult:
    def __init__(self, commands: List[str], exit_codes: List[int], stdout: str = "", stderr: str = "",
                 hide: bool = False) -> None:
        self.commands = commands
        self.stdout = stdout
        self.stderr = stderr
        self.exit_codes = exit_codes
        self.hide = hide

    @property
    def command(self) -> str:
        if len(self.commands) > 1:
            raise ValueError("Commands cannot be merged")

        return next(iter(self.commands))

    @property
    def exited(self) -> int:
        if len(self.exit_codes) > 1:
            raise ValueError("Exit codes cannot be merged")

        return next(iter(self.exit_codes))

    @property
    def return_code(self) -> int:
        return self.exited

    @property
    def ok(self) -> bool:
        return all(code == 0 for code in self.exit_codes)

    @property
    def failed(self) -> bool:
        return not self.ok

    def __bool__(self) -> bool:
        return self.ok

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, RunnersResult):
            return False

        return self.exited == other.exited \
            and self.stdout == other.stdout \
            and self.stderr == other.stderr

    def __ne__(self, other: object) -> bool:
        return not self == other

    @staticmethod
    def merge(results: List['RunnersResult']) -> 'RunnersResult':
        len_ = len(results)
        if len_ == 0:
            raise ValueError("At least one result should be present")
        elif len_ == 1:
            return results[0]

        hide = results[0].hide
        commands = []
        stdout = ""
        stderr = ""
        exit_codes = []
        for result in results:
            if result.hide != hide:
                raise ValueError("Cannot merge instances of RunnersResult with different 'hide' property")
            commands.extend(result.commands)
            stdout += result.stdout
            stderr += result.stderr
            exit_codes.extend(result.exit_codes)

        return RunnersResult(commands, exit_codes, stdout, stderr, hide=hide)

    def __str__(self) -> str:
        return f"Command exited with status {self.repr_code()}.\n" \
               f"{self.repr_out()}"

    def repr_code(self) -> str:
        unique_codes = set(self.exit_codes)
        return str(self.exit_codes[0]) if len(unique_codes) == 1 else "'undefined'"

    def repr_out(self, *, hide_already_printed: bool = False) -> str:
        ret = []
        for x in ("stdout", "stderr"):
            val: str = getattr(self, x)
            if val:
                val = 'already printed' if hide_already_printed and not self.hide else val.rstrip()
                ret.append(f"=== {x} ===\n"
                           f"{val.rstrip()}\n")
        return "\n".join(ret)

    def grep_returned_nothing(self) -> bool:
        return not self.stdout and not self.stderr and self.exited == 1


class Failure(Exception):
    def __init__(self, result: RunnersResult):
        self.result = result

    def __str__(self) -> str:
        ret = self._header()
        repr_out = self.result.repr_out(hide_already_printed=True)
        if repr_out:
            ret.append(repr_out)

        return "\n".join(ret)

    def _header(self) -> List[str]:
        return []


class CommandInterrupted(Failure, KeyboardInterrupt):
    def _header(self) -> List[str]:
        return [
            "\n",
            f"Command: {self.result.command!r}\n",
            f"Exit code: {self.result.exited}\n",
        ]


class UnexpectedExit(Failure):
    def _header(self) -> List[str]:
        return [
            "Encountered a bad command exit code!\n",
            f"Command: {self.result.command!r}\n",
            f"Exit code: {self.result.exited}\n",
        ]


class CommandTimedOut(Failure):
    def __init__(self, result: RunnersResult, timeout: int):
        super().__init__(result)
        self.timeout = timeout

    def _header(self) -> List[str]:
        return [
            f"Command did not complete within {self.timeout} seconds!\n",
            f"Command: {self.result.command!r}\n",
        ]


Token = int
GenericResult = Union[BaseException, RunnersResult, fabric.transfer.Result]
HostToResult = Dict[str, GenericResult]
TokenizedResult = OrderedDict[Token, GenericResult]

RESULT = TypeVar('RESULT', bound=GenericResult, covariant=True)


class GroupResult(Generic[RESULT]):
    def __init__(self, results: Mapping[str, Sequence[RESULT]]) -> None:
        self._results = results

    def __str__(self) -> str:
        return represent_results(self._results)


class BaseExecutorException(BaseException):
    def __init__(self, results: Mapping[str, Sequence[GenericResult]]):
        self.results = results

    def __str__(self) -> str:
        # We always print output of all commands in the batch even if they are not failed,
        # for the reason that the user code might want to print output of some commands in the batch,
        # but failed to do that because of the exception.
        return represent_results(self.results, hide_already_printed=True)


class GroupInterrupt(BaseExecutorException, KeyboardInterrupt):
    def __str__(self) -> str:
        return '\n' + super().__str__()


class GroupException(BaseExecutorException, errors.GroupException):
    pass


class Callback(ABC):
    """
    Callback to process the result of commands.
    """

    @abstractmethod
    def accept(self, host: str, token: Token, result: RunnersResult) -> None:
        """
        The method is called after the run / sudo command is exited.
        Calling of the method happens sequentially in one thread after some batch of commands is executed on all nodes.

        For the particular host, the order of results with which the method is called
        corresponds to the order of queued commands, for which the given callback was requested.

        :param host: host on which the command was executed
        :param token: identifier of the queued action
        :param result: RunnersResult instance holding the result of the exited command.
        """
        pass


_RawHostToResult = Dict[str, Union[BaseException, fabric.runners.Result, fabric.transfer.Result]]

_Action = Tuple[str, tuple, dict]
_PayloadItem = Tuple[_Action, Optional[Callback], Token]

_T = TypeVar('_T', bound='RawExecutor')


class RawExecutor:

    def __init__(self, cluster: Environment, connection_pool: ConnectionPool = None) -> None:
        self.logger = cluster.log
        if connection_pool is not None:
            self.connection_pool = connection_pool
        else:
            self.connection_pool = cluster.connection_pool
        self._connections_queue: Dict[str, List[_PayloadItem]] = {}
        self._last_token = -1
        self._last_results: Dict[str, TokenizedResult] = {}
        self._command_separator = ''.join(random.choice('=-_') for _ in range(32))
        self._supported_args = {'hide', 'warn', 'pty', 'timeout', 'env', 'out_stream', 'err_stream'}
        self._interrupt_queue = threading.Semaphore(0)
        self._closed = False

    def __enter__(self: _T) -> _T:
        self._check_closed()
        return self

    def __exit__(self, exc_type: Optional[Type[BaseException]], exc_value: Optional[BaseException],
                 tb: Optional[TracebackType]) -> None:
        if not isinstance(exc_value, KeyboardInterrupt) and self._connections_queue:
            self.flush()

        self._closed = True

    def _check_closed(self) -> None:
        if self._closed:
            raise ValueError("Executor is closed")

    def _actions_mergeable(self, action1: _Action, action2: _Action) -> bool:
        do_type1, _, kwargs1 = action1
        do_type2, _, kwargs2 = action2
        if do_type1 not in ["sudo", "run"] or do_type1 != do_type2:
            return False

        for arg in self._supported_args:
            if arg in ('out_stream', 'err_stream') and kwargs1.get(arg) is kwargs2.get(arg):
                continue

            if arg in ('hide', 'warn', 'pty') and kwargs1.get(arg) == kwargs2.get(arg):
                continue

            # If any action has 'env' param, they are not mergeable.
            # Actions that have specific 'timeout' params are also not mergeable,
            # as we need to apply the specified timeout to the exactly one command.
            if kwargs1.get(arg) is not None or kwargs2.get(arg) is not None:
                return False

        return True

    def _reparse_results(self, raw_results: _RawHostToResult,
                         batch: Dict[str, List[_PayloadItem]]) -> Dict[str, TokenizedResult]:
        reparsed_results: Dict[str, TokenizedResult] = {}
        for host, raw_result in raw_results.items():
            payloads = batch[host]
            conn_results: TokenizedResult = collections.OrderedDict()
            reparsed_results[host] = conn_results

            runner_exception = None
            if isinstance(raw_result, (invoke.UnexpectedExit, invoke.CommandTimedOut, connections.CommandInterrupted)):
                runner_exception = raw_result
                raw_result = raw_result.result

            if not isinstance(raw_result, fabric.runners.Result):
                token = payloads[0][2]
                conn_results[token] = raw_result
                continue

            results = self._reparse_fabric_result(payloads, raw_result)

            for i, result in enumerate(results):
                reparsed_result: GenericResult = result
                _, callback, token = payloads[i]
                if i == len(results) - 1 and runner_exception is not None:
                    if isinstance(runner_exception, connections.CommandInterrupted):
                        # Part of commands were successful until the last reparsed command interrupted.
                        reparsed_result = CommandInterrupted(result)
                    elif isinstance(runner_exception, invoke.UnexpectedExit):
                        # Commands were successful until the last command in the batch.
                        reparsed_result = UnexpectedExit(result)
                    if isinstance(runner_exception, invoke.CommandTimedOut):
                        # Commands were successful until the last reparsed command.

                        # Take common timeout as timeout for the last command.
                        # This is not honest enough, as previous commands also consumed some time.
                        # This is acceptable for an exceptional case when we did not expect long-running command,
                        # as this is applicable only for commands without explicit timeout
                        # (i.e. having timeout=globals.nodes.command_execution.timeout).
                        reparsed_result = CommandTimedOut(result, runner_exception.timeout)

                conn_results[token] = reparsed_result
                if callback is not None and isinstance(reparsed_result, RunnersResult):
                    callback.accept(host, token, reparsed_result)

        return reparsed_results

    def _reparse_fabric_result(self, payloads: List[_PayloadItem],
                               result: fabric.runners.Result) -> List[RunnersResult]:
        # unpack last action in list of payloads
        _, _, kwargs = payloads[-1][0]
        hide = kwargs.get('hide', False)
        pty = kwargs.get('pty', False)

        # Raw stdout may not fully contain the separator,
        # if the channel was interrupted or closed due to a timeout, while reading the separator from the remote output.
        # In this case the part of the separator becomes a part of the reparsed command output.
        # This is acceptable as the situation is considered extremely rare,
        # and because the only we can do with not finished command is to print its partial output.

        # stdout and stderr are combined if pty.
        raw_stderrs = [] if pty else result.stderr.split(self._command_separator)
        raw_stdouts = result.stdout.split(self._command_separator)

        results = []
        result_i = 0
        stdout = stderr = ''
        code = result.exited

        out_i = 0
        err_i = 0
        has_next = True
        while has_next:
            has_next = out_i < len(raw_stdouts) and (pty or err_i < len(raw_stderrs))
            if has_next:
                stdout = raw_stdouts[out_i]
                if not pty:
                    stderr = raw_stderrs[err_i]
                if result_i > 0:
                    # cut LF from the previous "echo <separator>" command
                    stdout = stdout[1:]
                    if not pty:
                        stderr = stderr[1:]

                has_next = out_i + 1 < len(raw_stdouts)
                if has_next:
                    matcher = EXIT_CODE_PATTERN.match(raw_stdouts[out_i + 1])
                    if matcher is not None:
                        code = int(matcher.group(1))
                    else:
                        has_next = False

                has_next = has_next and (pty or err_i + 1 < len(raw_stderrs))

            action, _, _ = payloads[result_i]
            command: str = action[1][0]
            results.append(RunnersResult([command], [code], stdout, stderr, hide=hide))

            stdout = stderr = ''
            code = result.exited
            result_i += 1

            err_i += 1
            out_i += 2

        return results

    def _get_separator(self, kwargs: dict) -> str:
        warn: bool = kwargs.get('warn', False)
        pty: bool = kwargs.get('pty', False)
        if warn:
            separator_symbol = ";"
        else:
            separator_symbol = "&&"

        separator = (
            f" {separator_symbol} "
            f"printf \"%s\\n$?\\n%s\\n\" \"{self._command_separator}\" \"{self._command_separator}\" {separator_symbol} ")
        if not pty:
            separator += f"echo \"{self._command_separator}\" 1>&2 {separator_symbol} "

        return separator

    def _merge_actions(self, payload_items: List[_PayloadItem]) -> List[List[_PayloadItem]]:
        merged_payloads: List[List[_PayloadItem]] = []

        for payload in payload_items:
            if not merged_payloads:
                merged_payloads.append([payload])
                continue

            action = payload[0]
            previous_payloads = merged_payloads[-1]
            previous_action = previous_payloads[-1][0]
            if not self._actions_mergeable(previous_action, action):
                merged_payloads.append(([payload]))
                continue

            previous_payloads.append(payload)

        return merged_payloads

    def _get_callables(self) -> List[Dict[str, List[_PayloadItem]]]:
        callables: Dict[str, List[List[_PayloadItem]]] = {}

        for host, payload_items in self._connections_queue.items():
            callables[host] = self._merge_actions(payload_items)

        i = 0
        batches: List[Dict[str, List[_PayloadItem]]] = []

        while True:
            batch: Dict[str, List[_PayloadItem]] = {}
            for host, actions in callables.items():
                if len(actions) > i:
                    batch[host] = actions[i]
            if not batch:
                break

            i += 1
            batches.append(batch)

        return batches

    def _next_token(self) -> int:
        self._last_token += 1
        return self._last_token

    def queue(self, target: List[str], action: _Action, callback: Callback = None) -> int:
        self._check_closed()
        token = self._next_token()

        do_type, args, kwargs = action
        repr_args = self._repr_args(do_type, args)
        not_supported_args = set(kwargs.keys()) - self._supported_args
        if not_supported_args:
            raise Exception(f"Arguments {', '.join(map(repr, not_supported_args))} are not supported")

        if not target:
            self.logger.verbose('No nodes to perform %s %s with options: %s' % (do_type, repr_args, kwargs))
        else:
            self.logger.verbose(
                'Performing %s %s on nodes %s with options: %s' % (do_type, repr_args, list(target), kwargs))
            for host in target:
                self._connections_queue.setdefault(host, []).append((action, callback, token))

        return token

    def get_last_results(self) -> Dict[str, TokenizedResult]:
        return self._last_results

    def get_flat_result(self) -> Mapping[str, Sequence[GenericResult]]:
        return get_flat_result(self._last_results)

    def flush(self) -> None:
        """
        Flushes the connections' queue and returns grouped result

        :return: grouped tokenized results per connection.
        """
        self._check_closed()
        self._last_results.clear()

        if not self._connections_queue:
            self.logger.verbose('Queue is empty, nothing to perform')
            return

        callable_batches: List[Dict[str, List[_PayloadItem]]] = self._get_callables()

        max_workers = len(self._connections_queue)

        with ThreadPoolExecutor(max_workers=max_workers) as TPE:
            for batch in callable_batches:
                # filter out hosts with failed commands
                batch = {host: payloads for host, payloads in batch.items()
                         # failed command is always last if present
                         if host not in self._last_results
                         or not isinstance(list(self._last_results[host].values())[-1], BaseException)}

                retry = 0
                while True:
                    retry += 1

                    self._do_batch(batch, TPE, self._last_results)
                    batch = self._get_remained_batch(batch)

                    if (not batch or retry >= static.GLOBALS['workaround']['retries']
                            or not self._try_workaround(batch, TPE)):
                        break

                    self.logger.verbose('Retrying #%s...' % retry)
                    time.sleep(static.GLOBALS['workaround']['delay_period'])

        self._connections_queue = {}

        for results in self._last_results.values():
            if any(isinstance(result, BaseException) for result in results.values()):
                raise GroupException(self.get_flat_result())

    def interrupt(self) -> None:
        self._interrupt_queue.release()

    def _prepare_merged_action(self, host: str, payloads: List[_PayloadItem]) -> _Action:
        # unpack last action in list of payloads
        do_type, args, kwargs = payloads[-1][0]

        if do_type == 'get':
            self.logger.verbose('Executing get %s on host %s with options: %s' % (args, host, kwargs))
        if do_type == 'put':
            self.logger.verbose(
                'Executing put %s on host %s with options: %s' % (self._repr_args(do_type, args), host, kwargs))
            if isinstance(args[0], bytes):
                # Each thread should use its own instance of BytesIO.
                args = io.BytesIO(args[0]), args[1]

        if do_type in ('run', 'sudo'):
            if kwargs.get('timeout', None) is None:
                kwargs = {**kwargs, 'timeout': static.GLOBALS['nodes']['command_execution']['timeout']}

            commands: List[str] = [action[1][0] for action, _, _ in payloads]
            self.logger.verbose('Executing %s %s on host %s with options: %s' % (do_type, commands, host, kwargs))

            precommand = ''
            if do_type == 'sudo':
                precommand = 'sudo '

            separator = self._get_separator(kwargs)

            merged_command = (separator + precommand).join(commands)
            args = (merged_command,)

        return do_type, args, kwargs

    @staticmethod
    def _repr_args(do_type: str, args: tuple) -> tuple:
        if do_type == 'put' and isinstance(args[0], bytes):
            return '<text>', args[1]

        return args

    def _flush_logger_writers(self, batch: Dict[str, List[_PayloadItem]]) -> None:
        for payloads in batch.values():
            # Actions are merged only if there are no out_stream/err_stream or if they are the same instance.
            # It is thus enough to take only first action from payloads
            payload = payloads[0]
            action, _, _ = payload
            _, _, kwargs = action
            for stream_key in ('out_stream', 'err_stream'):
                if isinstance(kwargs.get(stream_key), log.LoggerWriter):
                    writer: log.LoggerWriter = kwargs[stream_key]
                    writer.flush(remainder=True)

    def _do_batch(self, batch: Dict[str, List[_PayloadItem]], tpe: ThreadPoolExecutor,
                  capture_results: Dict[str, TokenizedResult]) -> None:
        results: _RawHostToResult = {}
        futures: Dict[str, concurrent.futures.Future] = {}

        def safe_exec(result_map: Dict[str, Any], host: str, call: Callable[[], Any]) -> None:
            try:
                result_map[host] = call()
            except BaseException as e:  # including KeyboardInterrupt
                results[host] = e

        interrupted = False
        try:
            while True:
                try:
                    if not interrupted:
                        for host, payloads in batch.items():
                            cxn = self.connection_pool.get_connection(host)
                            cxn.KM_start()

                            do_type, args, kwargs = self._prepare_merged_action(host, payloads)
                            # pylint: disable-next=cell-var-from-loop
                            safe_exec(futures, host, lambda: tpe.submit(getattr(cxn, do_type), *args, **kwargs))

                    # Timeout is implemented through timeout for run/sudo that finishes the future eventually.
                    # For put/get fabric & paramiko do not offer timeout, so transfer can be stopped only using SIGINT.
                    # The following short-time sleeps allow to implement interruption.
                    not_done = set(futures.values())
                    while not_done := concurrent.futures.wait(not_done, timeout=connections.input_sleep).not_done:
                        if self._interrupt_queue.acquire(blocking=False):  # pylint: disable=consider-using-with
                            raise KeyboardInterrupt()

                except KeyboardInterrupt:
                    interrupted = True
                    for host, payloads in batch.items():
                        cxn = self.connection_pool.get_connection(host)
                        cxn.KM_interrupt()

                    continue

                break
        finally:
            for host, future in futures.items():
                # Only obtain the result of finished futures, thus no timeout
                # pylint: disable-next=cell-var-from-loop
                safe_exec(results, host, lambda: future.result(timeout=None))

        self._flush_logger_writers(batch)

        parsed_results = self._reparse_results(results, batch)
        for host, tokenized_results in parsed_results.items():
            capture_results.setdefault(host, collections.OrderedDict()).update(tokenized_results)

        if interrupted:
            flat_results = get_flat_result(capture_results)
            timed_out = False
            # Check if at least one command was really interrupted.
            # See also `connections.RemoteRunner.wait()`.
            for flat_result in flat_results.values():
                for result in flat_result:
                    if isinstance(result, KeyboardInterrupt):
                        raise GroupInterrupt(flat_results)

                    timed_out = timed_out or isinstance(result, CommandTimedOut)

            # No command was interrupted using real SIGINT character sent, but some command was timed out.
            # This case will be processed later.
            if timed_out:
                return

            # The commands were really finished, but we still need to stop execution of main thread.
            # It seems that no need to print the output.
            raise KeyboardInterrupt()

    def _get_remained_batch(self, batch: Dict[str, List[_PayloadItem]]) -> Dict[str, List[_PayloadItem]]:
        remained_batch = {}
        for host, payloads in batch.items():
            remained_payloads = []
            for payload in payloads:
                _, _, token = payload
                tokenized_results = self._last_results[host]
                # Command is not yet executed or failed.
                # Not yet executed commands can be only after the failed command.
                if token not in tokenized_results or isinstance(tokenized_results[token], BaseException):
                    remained_payloads.append(payload)

            if remained_payloads:
                remained_batch[host] = remained_payloads

        return remained_batch

    def _try_workaround(self, batch: Dict[str, List[_PayloadItem]], tpe: ThreadPoolExecutor) -> bool:
        not_booted = []

        for host in batch:
            # failed command is always last
            exception = list(self._last_results[host].values())[-1]
            if isinstance(exception, CommandTimedOut):
                self.logger.verbose("Command timed out at %s: %s" % (host, str(exception.result)))
                return False
            elif isinstance(exception, UnexpectedExit):
                # Do not str(exception) because it discards output in case of hide=False
                exception_message = str(exception.result)
            else:
                exception_message = str(exception)

            if self._is_allowed_etcd_exception(exception_message):
                self.logger.verbose("Detected ETCD problem at %s, need retry: %s" % (host, exception_message))
            elif self._is_allowed_kubernetes_exception(exception_message):
                self.logger.verbose("Detected kubernetes problem at %s, need retry: %s" % (host, exception_message))
            elif self._is_allowed_connection_exception(exception_message):
                self.logger.verbose("Detected connection exception at %s, will try to reconnect to node. Exception: %s"
                                    % (host, exception_message))
                not_booted.append(host)
            else:
                self.logger.verbose("Detected unavoidable exception at %s, trying to solve automatically: %s"
                                         % (host, exception_message))
                return False

        if not_booted:
            results = self._wait_for_boot_with_executor(not_booted, tpe)
            # if there are not booted nodes, but we succeeded to wait for at least one is booted,
            # we can continue execution
            if all(isinstance(result, BaseException) for result in results.values()):
                return False

        return True

    def wait_for_boot(self, left_nodes: List[str], timeout: int = None,
                      initial_boot_history: Mapping[str, RunnersResult] = None) -> HostToResult:
        with ThreadPoolExecutor(max_workers=len(left_nodes)) as TPE:
            return self._wait_for_boot_with_executor(left_nodes, TPE, timeout, initial_boot_history)

    def _wait_for_boot_with_executor(self, left_nodes: List[str], tpe: ThreadPoolExecutor,
                                     overridden_timeout: int = None,
                                     initial_boot_history: Mapping[str, RunnersResult] = None) -> HostToResult:

        timeout = self._resolve_boot_timeout(left_nodes, overridden_timeout)
        delay_period = static.GLOBALS['nodes']['boot']['defaults']['delay_period']

        if initial_boot_history is None:
            initial_boot_history = {}

        results: HostToResult = {}
        time_start = datetime.now()

        self.logger.verbose("Trying to connect to nodes, timeout is %s seconds..." % timeout)

        # each connection has timeout, so the only we need is to repeat connecting attempts
        # during specified number of seconds
        while True:
            attempt_time_start = datetime.now()
            self._disconnect(left_nodes)

            self.logger.verbose("Attempting to connect to nodes...")
            # this should be invoked without explicit timeout, and relied on fabric Connection timeout instead.
            results.update(self._do_nopasswd(left_nodes, tpe, "last reboot"))
            left_nodes = [host for host, result in results.items()
                          if (isinstance(result, BaseException)
                              # Something is wrong with sudo access. Node is active.
                              and not self.is_require_nopasswd_exception(result)
                              # If not a connection-related exception, do not try to connect further
                              and self._is_allowed_connection_exception(str(result)))
                          or (isinstance(result, RunnersResult)
                              and result == initial_boot_history.get(host))]

            waited = (datetime.now() - time_start).total_seconds()
            if left_nodes:
                timeout = self._resolve_boot_timeout(left_nodes, overridden_timeout)

            if not left_nodes or waited >= timeout:
                break

            self.logger.verbose("Nodes %s are not ready yet, remaining time to wait %i"
                                % (left_nodes, timeout - waited))

            attempt_time = (datetime.now() - attempt_time_start).total_seconds()
            if attempt_time < delay_period:
                time.sleep(delay_period - attempt_time)

        if left_nodes:
            self.logger.verbose("Failed to wait for boot of nodes %s" % left_nodes)
        else:
            self.logger.verbose("All nodes are online now")

        return results

    def _resolve_boot_timeout(self, hosts: List[str], overridden_timeout: Optional[int]) -> int:
        if overridden_timeout is not None:
            return overridden_timeout

        return max(self._get_node_boot_timeout(host) for host in hosts)

    def _get_node_boot_timeout(self, host: str) -> int:
        timeout: int = self.connection_pool.get_node(host)['boot']['timeout']
        return timeout

    def _do_nopasswd(self, left_nodes: List[str], tpe: ThreadPoolExecutor, command: str) -> HostToResult:
        prompt = '[sudo] password: '

        class NoPasswdResponder(invoke.Responder):
            def __init__(self) -> None:
                super().__init__(re.escape(prompt), "")

            def submit(self, stream: str) -> Generator[str, None, None]:
                if self.pattern_matches(stream, self.pattern, "index"):
                    # If user appears to be not a NOPASSWD sudoer, "sudo" suggests to write password.
                    # This is a W/A to handle the situation in a docker container without pseudo-TTY (no -t option)
                    # As long as we require NOPASSWD, we can just fail immediately in such cases.
                    raise invoke.exceptions.ResponseNotAccepted("The user should be a NOPASSWD sudoer")

                # The only acceptable situation, responder does nothing.
                yield from []

        # Currently only NOPASSWD sudoers are supported.
        # Thus, running of connection.sudo("something") should be equal to connection.run("sudo something")
        token = self._next_token()
        action: _Action = ("run", (f"sudo -S -p '{prompt}' {command}",),
                           {"hide": True, "pty": True, "watchers": [NoPasswdResponder()]})
        payload: _PayloadItem = (action, None, token)
        batch = {host: [payload] for host in left_nodes}
        parsed_results: Dict[str, TokenizedResult] = {}
        self._do_batch(batch, tpe, parsed_results)
        return {host: next(iter(results.values())) for host, results in parsed_results.items()}

    @staticmethod
    def is_require_nopasswd_exception(exc: BaseException) -> bool:
        return isinstance(exc, invoke.exceptions.Failure) \
               and isinstance(exc.reason, invoke.exceptions.ResponseNotAccepted)

    @staticmethod
    def _is_allowed_connection_exception(exception_message: str) -> bool:
        exception_message = exception_message.partition('\n')[0]
        for known_exception_message in static.GLOBALS['connection']['bad_connection_exceptions']:
            if known_exception_message in exception_message:
                return True

        return False

    @staticmethod
    def _is_allowed_etcd_exception(exception_message: str) -> bool:
        for known_exception_message in static.GLOBALS['etcd']['temporary_exceptions']:
            if known_exception_message in exception_message:
                return True

        return False

    @staticmethod
    def _is_allowed_kubernetes_exception(exception_message: str) -> bool:
        for known_exception_message in static.GLOBALS['kubernetes']['temporary_exceptions']:
            if known_exception_message in exception_message:
                return True

        return False

    def _disconnect(self, hosts: List[str]) -> None:
        for host in hosts:
            self.logger.verbose('Disconnected session with %s' % host)
            cxn = self.connection_pool.get_connection(host)
            cxn.close()


def get_flat_result(results: Dict[str, TokenizedResult]) -> Mapping[str, Sequence[GenericResult]]:
    return {host: list(res.values()) for host, res in results.items()}


def represent_results(flat_results: Mapping[str, Sequence[GenericResult]], hide_already_printed: bool = True) -> str:
    host_outputs = []
    for host, results in flat_results.items():
        output = f"{host}:"

        # filter out transfer results and the last exception if present
        runners_results = [res for res in results if isinstance(res, RunnersResult)]
        if runners_results:
            merged_result = RunnersResult.merge(runners_results)
            output += f" code={merged_result.repr_code()}"
            repr_out = merged_result.repr_out(hide_already_printed=hide_already_printed)
            if repr_out:
                output += '\n\t' + repr_out.replace('\n', '\n\t')

        # The exception may be only the last in the list. We are also sure to have at least one result per node.
        if isinstance(results[-1], BaseException):
            exception = errors.wrap_kme_exception(results[-1])
            output += '\n\t' + str(exception).replace('\n', '\n\t')

        host_outputs.append(output)

    return "\n".join(host_outputs)



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import math
import os
import time
from contextlib import contextmanager
from typing import List, Dict, Iterator, Any, Optional

import yaml
from jinja2 import Template
from ordered_set import OrderedSet

from kubemarine import system, admission, etcd, packages, jinja, sysctl
from kubemarine.core import utils, static, summary, log, errors
from kubemarine.core.cluster import KubernetesCluster, EnrichmentStage, enrichment
from kubemarine.core.executor import Token
from kubemarine.core.group import NodeGroup, DeferredGroup, RunnersGroupResult, CollectorCallback
from kubemarine.core.errors import KME
from kubemarine.core.yaml_merger import default_merger
from kubemarine.cri import containerd
from kubemarine.kubernetes import components

ERROR_DOWNGRADE='Kubernetes old version \"%s\" is greater than new one \"%s\"'
ERROR_SAME='Kubernetes old version \"%s\" is the same as new one \"%s\"'
ERROR_MAJOR_RANGE_EXCEEDED='Major version \"%s\" rises to new \"%s\" more than one'
ERROR_MINOR_RANGE_EXCEEDED='Minor version \"%s\" rises to new \"%s\" more than one'
ERROR_NOT_LATEST_PATCH='New version \"%s\" is not the latest supported patch version \"%s\"'

ERROR_KUBELET_PATCH_NOT_KUBERNETES_NODE = "%s patch can be uploaded only to control-plane or worker nodes"
ERROR_CONTROL_PLANE_PATCH_NOT_CONTROL_PLANE_NODE = "%s patch can be uploaded only to control-plane nodes"

ERROR_UPGRADE_UNEXPECTED_PROPERTY='Unexpected %s properties in the procedure inventory for upgrade.'

ERROR_AMBIGUOUS_CONNTRACK_MAX = "Detected ambiguous 'net.netfilter.nf_conntrack_max' value: {values}"


@enrichment(EnrichmentStage.PROCEDURE, procedures=['upgrade'])
def enrich_upgrade_inventory(cluster: KubernetesCluster) -> None:
    procedure_inventory = cluster.procedure_inventory
    allowed_properties = {
        'upgrade_plan', 'upgrade_nodes', 'disable-eviction', 'prepull_group_size', 'grace_period', 'drain_timeout'
    }
    allowed_properties.update(procedure_inventory['upgrade_plan'])
    unexpected_properties = set(procedure_inventory) - allowed_properties
    if unexpected_properties:
        raise Exception(ERROR_UPGRADE_UNEXPECTED_PROPERTY % (', '.join(map(repr, unexpected_properties)),))

    upgrade_version = get_procedure_upgrade_version(cluster)
    cluster.inventory.setdefault("services", {}).setdefault("kubeadm", {})['kubernetesVersion'] = upgrade_version


@enrichment(EnrichmentStage.PROCEDURE, procedures=['upgrade'])
def verify_upgrade_inventory(cluster: KubernetesCluster) -> None:
    initial_kubernetes_version = get_kubernetes_version(cluster.previous_inventory)
    upgrade_version = get_kubernetes_version(cluster.inventory)

    test_version_upgrade_possible(initial_kubernetes_version, upgrade_version)

    cluster.log.info(
        '------------------------------------------\nUPGRADING KUBERNETES %s ⭢ %s\n------------------------------------------' % (
        initial_kubernetes_version, upgrade_version))

    dump_directory = utils.get_dump_directory(cluster.context)
    if jinja.is_template(get_procedure_upgrade_version(cluster)) and os.path.exists(dump_directory):
        os.rename(dump_directory, os.path.join(os.path.dirname(dump_directory), upgrade_version))
        cluster.context['dump_subdir'] = upgrade_version


@enrichment(EnrichmentStage.PROCEDURE, procedures=['restore'])
def enrich_restore_inventory(cluster: KubernetesCluster) -> None:
    logger = cluster.log
    inventory = cluster.inventory
    backup_version = cluster.context['backup_descriptor'].get('kubernetes', {}).get('version')
    if not backup_version:
        logger.warning("Not possible to verify Kubernetes version, as descriptor does not contain 'kubernetes.version'")
        return

    installed_version = get_kubernetes_version(inventory)
    if backup_version != installed_version and not jinja.is_template(installed_version):
        logger.warning(f'Installed kubernetes version {installed_version} '
                       f'does not match version from backup {backup_version}')

    inventory.setdefault("services", {}).setdefault("kubeadm", {})['kubernetesVersion'] = backup_version


@enrichment(EnrichmentStage.PROCEDURE, procedures=['reconfigure'])
def enrich_reconfigure_inventory(cluster: KubernetesCluster) -> None:
    kubeadm_sections = utils.subdict_yaml(
        cluster.procedure_inventory.get('services', {}),
        ['kubeadm', 'kubeadm_kubelet', 'kubeadm_kube-proxy', 'kubeadm_patches'])

    if kubeadm_sections:
        default_merger.merge(cluster.inventory.setdefault('services', {}), utils.deepcopy_yaml(kubeadm_sections))


@enrichment(EnrichmentStage.ALL)
def verify_roles(cluster: KubernetesCluster) -> None:
    control_plane_roles = ['control-plane']
    if cluster.context['initial_procedure'] == 'do':
        control_plane_roles = ['control-plane', 'master']

    if cluster.make_group_from_roles(control_plane_roles).is_empty():
        raise KME("KME0004")


@enrichment(EnrichmentStage.FULL)
def enrich_inventory(cluster: KubernetesCluster) -> None:
    inventory = cluster.inventory
    kubeadm = inventory['services']['kubeadm']
    kubeadm['dns'].setdefault('imageRepository', f"{kubeadm['imageRepository']}/coredns")

    enriched_certsans = []

    for node in inventory["nodes"]:
        if 'balancer' in node['roles'] or 'control-plane' in node['roles']:
            enriched_certsans.extend([node['name'], node['internal_address']])
            if node.get('address') is not None:
                enriched_certsans.append(node['address'])

    # The VRRP IP may be actually unused, but let's add it because it is probably specified to be used in the future.
    for item in inventory["vrrp_ips"]:
        enriched_certsans.append(item['ip'])
        if item.get("floating_ip"):
            enriched_certsans.append(item["floating_ip"])

    if inventory.get("public_cluster_ip"):
        enriched_certsans.append(inventory["public_cluster_ip"])

    certsans = kubeadm['apiServer']['certSANs']

    # do not overwrite apiServer.certSANs, but append - may be user specified something already there?
    for name in enriched_certsans:
        if name not in certsans:
            certsans.append(name)

    # validating node labels and configuring additional labels
    for node in inventory["nodes"]:
        if "control-plane" not in node["roles"] and "worker" not in node["roles"]:
            if "labels" in node:
                raise Exception("Only 'worker' or 'control-plane' nodes can have labels, "
                                "but found label on %s, roles: %s" % (node["name"], node["roles"]))
            if "taints" in node:
                raise Exception("Only 'worker' or 'control-plane' nodes can have taints, "
                                "but found taints on %s, roles: %s" % (node["name"], node["roles"]))
            continue

        if "worker" in node["roles"]:
            if "labels" not in node:
                node["labels"] = {}
            node["labels"]["node-role.kubernetes.io/worker"] = "worker"

    # Validate the provided podSubnet and serviceSubnet IP addresses
    for subnet in ('podSubnet', 'serviceSubnet'):
        utils.isipv(kubeadm['networking'][subnet], [4, 6])

    # validate nodes in kubeadm_patches (groups are validated with JSON schema)
    for control_plane_item, patches in inventory["services"]["kubeadm_patches"].items():
        for patch in patches:
            if 'nodes' not in patch:
                continue

            for node in cluster.get_nodes_by_names(patch['nodes']):
                if control_plane_item == 'kubelet' and 'control-plane' not in node['roles'] and 'worker' not in node['roles']:
                    raise Exception(ERROR_KUBELET_PATCH_NOT_KUBERNETES_NODE % control_plane_item)
                if control_plane_item != 'kubelet' and ('control-plane' not in node['roles']):
                    raise Exception(ERROR_CONTROL_PLANE_PATCH_NOT_CONTROL_PLANE_NODE % control_plane_item)

    # check ignorePreflightErrors value and add mandatory errors from defaults.yaml if they're absent
    default_preflight_errors = static.DEFAULTS["services"]["kubeadm_flags"]["ignorePreflightErrors"].split(",")
    preflight_errors = inventory["services"]["kubeadm_flags"]["ignorePreflightErrors"].split(",")

    preflight_errors.extend(default_preflight_errors)
    inventory["services"]["kubeadm_flags"]["ignorePreflightErrors"] = ",".join(set(preflight_errors))

    enrich_kube_proxy(cluster)


def enrich_kube_proxy(cluster: KubernetesCluster) -> None:
    inventory = cluster.inventory

    # override kubeadm_kube-proxy.conntrack.min with sysctl.net.netfilter.nf_conntrack_max
    # since they define the same kernel variable
    kubernetes_nodes = cluster.make_group_from_roles(['control-plane', 'worker'])
    conntrack_max_values = OrderedSet(
        sysctl.get_parameter(cluster, node, 'net.netfilter.nf_conntrack_max')
        for node in kubernetes_nodes.get_ordered_members_list()
    )

    if len(conntrack_max_values) > 1:
        raise Exception(ERROR_AMBIGUOUS_CONNTRACK_MAX.format(
            values='{' + ', '.join(map(repr, conntrack_max_values)) + '}'))

    conntrack_max = next(iter(conntrack_max_values), None)
    if components.kube_proxy_overwrites_higher_system_values(cluster) and conntrack_max is not None:
        inventory["services"]["kubeadm_kube-proxy"]["conntrack"]["min"] = conntrack_max
    else:
        inventory["services"]["kubeadm_kube-proxy"]["conntrack"].pop("min",None)


def reset_installation_env(group: NodeGroup) -> Optional[RunnersGroupResult]:
    log = group.cluster.log

    log.debug("Cleaning up previous installation...")

    cluster: KubernetesCluster = group.cluster
    procedure: str = cluster.context['initial_procedure']

    drain_timeout = cluster.procedure_inventory.get('drain_timeout')
    grace_period = cluster.procedure_inventory.get('grace_period')

    # if we perform "add" or "remove" node procedure
    # then we need to additionally perform "drain" and "delete" during reset
    nodes_for_draining = cluster.make_group([])

    # perform FULL reset only for "add" or "remove" procedures
    # do not perform full reset on cluster (re)installation, it could hang on last etcd member
    # nodes should be deleted only during "add" or "remove" procedures
    full_reset = procedure != 'install'

    nodes_for_manual_etcd_remove = cluster.make_group([])
    active_nodes = group.get_online_nodes(True)

    if procedure == 'remove_node':
        # We need to manually remove members from etcd for "remove" procedure,
        # only if corresponding nodes are not active.
        # Otherwise, they will be removed by "kubeadm reset" command.
        nodes_for_manual_etcd_remove = group.exclude_group(active_nodes)

        # kubectl drain command hands on till timeout is exceeded for nodes which are off
        # so we should drain only active nodes
        nodes_for_draining = active_nodes
    elif procedure == 'add_node':
        nodes_for_draining = group

    if not nodes_for_manual_etcd_remove.is_empty():
        log.warning(f"Nodes {nodes_for_manual_etcd_remove.get_hosts()} are considered as not active. "
                    "Full cleanup procedure cannot be performed. "
                    "Corresponding members will be removed from etcd manually.")
        etcd.remove_members(nodes_for_manual_etcd_remove)

    if not nodes_for_draining.is_empty():
        drain_nodes(nodes_for_draining, drain_timeout=drain_timeout, grace_period=grace_period)

    if full_reset and not active_nodes.is_empty():
        log.verbose(f"Resetting kubeadm on nodes {active_nodes.get_hosts()} ...")
        result = active_nodes.sudo('sudo kubeadm reset -f')
        log.debug("Kubeadm successfully reset:\n%s" % result)

    if not active_nodes.is_empty():
        log.verbose(f"Cleaning nodes {active_nodes.get_hosts()} ...")
        # bash semicolon mark will avoid script from exiting and will resume the execution
        result = active_nodes.sudo(
            'sudo kubeadm reset phase cleanup-node; '  # it is required to "cleanup-node" for all procedures
            'sudo systemctl stop kubelet; '
            'sudo rm -rf /etc/kubernetes/manifests /var/lib/kubelet/pki /var/lib/etcd /etc/kubernetes/patches; '
            'sudo mkdir -p /etc/kubernetes/manifests; ', warn=True, pty=True)

        # Disabled initial prune for images prepull feature. Need analysis for possible negative impact.
        # result.update(cri.prune(active_nodes, all_implementations=True))

        log.debug(f"Nodes {active_nodes.get_hosts()} cleaned up successfully:\n" + "%s" % result)

    if full_reset:
        return delete_nodes(group)

    return None


def drain_nodes(group: NodeGroup, disable_eviction: bool = False,
                drain_timeout: int = None, grace_period: int = None) -> RunnersGroupResult:
    cluster: KubernetesCluster = group.cluster
    log = cluster.log

    control_plane = cluster.get_unchanged_nodes().having_roles(['control-plane']).get_first_member()
    result = control_plane.sudo("kubectl get nodes -o custom-columns=NAME:.metadata.name")

    stdout = list(result.values())[0].stdout
    log.verbose("Detected the following nodes in cluster:\n%s" % stdout)

    for node in group.get_ordered_members_list():
        node_name = node.get_node_name()

        # Split stdout into lines
        stdout_lines = stdout.split('\n')[1:]
        # Check if node_name exactly matches any line
        if node_name in stdout_lines:
            log.debug("Draining node %s..." % node_name)
            drain_cmd = prepare_drain_command(
                cluster, node_name,
                disable_eviction=disable_eviction, drain_timeout=drain_timeout, grace_period=grace_period)
            control_plane.sudo(drain_cmd, hide=False, pty=True)
        else:
            log.warning("Node %s is not found in cluster and can't be drained" % node_name)

    return control_plane.sudo("kubectl get nodes")


def delete_nodes(group: NodeGroup) -> RunnersGroupResult:
    cluster: KubernetesCluster = group.cluster
    log = cluster.log

    control_plane = cluster.get_unchanged_nodes().having_roles(['control-plane']).get_first_member()
    result = control_plane.sudo("kubectl get nodes -o custom-columns=NAME:.metadata.name")

    stdout = list(result.values())[0].stdout
    log.verbose("Detected the following nodes in cluster:\n%s" % stdout)

    for node in group.get_ordered_members_list():
        node_name = node.get_node_name()
        
        # Split stdout into lines
        stdout_lines = stdout.split('\n')[1:]
        # Check if node_name exactly matches any line
        if node_name in stdout_lines:
            log.debug("Deleting node %s from the cluster..." % node_name)
            control_plane.sudo("kubectl delete node %s" % node_name, hide=False, pty=True)
        else:
            log.warning("Node %s is not found in cluster and can't be removed" % node_name)

    return control_plane.sudo("kubectl get nodes")


def install(group: NodeGroup) -> RunnersGroupResult:
    cluster: KubernetesCluster = group.cluster
    log = cluster.log

    with group.new_executor() as exe:
        log.debug("Making systemd unit...")
        for node in exe.group.get_ordered_members_list():
            node.sudo('rm -rf /etc/systemd/system/kubelet*')
            template = Template(utils.read_internal('templates/kubelet.service.j2')).render(
                hostname=node.get_node_name())
            log.debug("Uploading to '%s'..." % node.get_host())
            node.put(io.StringIO(template + "\n"), '/etc/systemd/system/kubelet.service', sudo=True)
            node.sudo("chmod 600 /etc/systemd/system/kubelet.service")

        log.debug("\nReloading systemd daemon...")
        system.reload_systemctl(exe.group)
        exe.group.sudo('systemctl enable kubelet')

    return group.sudo('systemctl status kubelet', warn=True)


def join_other_control_planes(group: NodeGroup) -> RunnersGroupResult:
    other_control_planes_group = group.get_ordered_members_list()[1:]

    join_dict = group.cluster.context["join_dict"]
    for node in other_control_planes_group:
        join_control_plane(group.cluster, node, join_dict)

    group.cluster.log.debug("Verifying installation...")
    first_control_plane = group.get_first_member()
    return first_control_plane.sudo("kubectl get pods --all-namespaces -o=wide")


def join_new_control_plane(group: NodeGroup) -> None:
    join_dict = get_join_dict(group)
    for node in group.get_ordered_members_list():
        join_control_plane(group.cluster, node, join_dict)


def join_control_plane(cluster: KubernetesCluster, node: NodeGroup, join_dict: dict) -> None:
    log = cluster.log
    node_name = node.get_node_name()

    join_config = components.get_init_config(cluster, node, init=False, join_dict=join_dict)
    config = components.get_kubeadm_config(cluster, join_config)

    utils.dump_file(cluster, config, 'join-config_%s.yaml' % node_name)

    log.debug("Uploading init config to control-plane '%s'..." % node_name)
    node.sudo("mkdir -p /etc/kubernetes")
    node.put(io.StringIO(config), '/etc/kubernetes/join-config.yaml', sudo=True)

    # put control-plane patches
    components.create_kubeadm_patches_for_node(cluster, node)

    # copy admission config to control-plane
    admission.copy_pss(node)

    # put audit-policy.yaml
    prepare_audit_policy(node)

    # ! ETCD on control-planes can't be initialized in async way, that is why it is necessary to disable async mode !
    log.debug('Joining control-plane \'%s\'...' % node_name)

    node.sudo(
        "kubeadm join "
        " --config=/etc/kubernetes/join-config.yaml "
        " --ignore-preflight-errors='" + cluster.inventory['services']['kubeadm_flags']['ignorePreflightErrors'] + "'"
        " --v=5",
        hide=False, pty=True)
    copy_admin_config(log, node)

    components.wait_for_pods(node)


@contextmanager
def local_admin_config(nodes: NodeGroup) -> Iterator[str]:
    temp_filepath = utils.get_remote_tmp_path()

    cluster_name = nodes.cluster.inventory['cluster_name']

    try:
        with nodes.new_executor() as exe:
            for defer in exe.group.get_ordered_members_list():
                internal_address = defer.get_config()['internal_address']
                if utils.isipv(internal_address, [6]):
                    internal_address = f"[{internal_address}]"

                defer.sudo(
                    f"cp /root/.kube/config {temp_filepath} "
                    f"&& sudo sed -i 's/{cluster_name}/{internal_address}/' {temp_filepath}")
        yield temp_filepath
    finally:
        nodes.sudo(f'rm -f {temp_filepath}')


def copy_admin_config(logger: log.EnhancedLogger, nodes: NodeGroup) -> None:
    logger.debug("Setting up admin-config...")
    command = "mkdir -p /root/.kube && sudo cp -f /etc/kubernetes/admin.conf /root/.kube/config"
    nodes.sudo(command)


def fetch_admin_config(cluster: KubernetesCluster) -> str:
    log = cluster.log

    first_control_plane = cluster.nodes['control-plane'].get_first_member()
    log.debug(f"Downloading kubeconfig from node {first_control_plane.get_node_name()!r}...")

    kubeconfig = list(first_control_plane.sudo('cat /root/.kube/config').values())[0].stdout

    # Replace cluster FQDN with ip
    public_cluster_ip = cluster.inventory.get('public_cluster_ip')
    if public_cluster_ip:
        if utils.isipv(public_cluster_ip, [6]):
            public_cluster_ip = f"[{public_cluster_ip}]"
        cluster_name = cluster.inventory['cluster_name']
        kubeconfig = kubeconfig.replace(cluster_name, public_cluster_ip)

    kubeconfig_filename = os.path.abspath("kubeconfig")
    utils.dump_file(cluster.context, kubeconfig, kubeconfig_filename, dump_location=False)
    cluster.log.debug(f"Kubeconfig saved to {kubeconfig_filename}")

    return kubeconfig_filename


def get_join_dict(group: NodeGroup) -> dict:
    cluster: KubernetesCluster = group.cluster
    first_control_plane = cluster.nodes["control-plane"].get_first_member()
    token_result = first_control_plane.sudo("kubeadm token create --print-join-command", hide=False)
    join_strings = list(token_result.values())[0].stdout.rstrip("\n")

    join_dict = {"worker_join_command": join_strings}
    join_array = join_strings[join_strings.find("--"):].split()
    for idx, _ in enumerate(join_array):
        current_string = join_array[idx]
        if "--" in current_string:
            join_dict[current_string.lstrip("--")] = join_array[idx + 1]

    cert_key_result = first_control_plane.sudo("kubeadm init phase upload-certs --upload-certs")
    cert_key = list(cert_key_result.values())[0].stdout.split("Using certificate key:\n")[1].rstrip("\n")
    join_dict["certificate-key"] = cert_key
    return join_dict


def init_first_control_plane(group: NodeGroup) -> None:
    cluster: KubernetesCluster = group.cluster
    log = cluster.log

    first_control_plane = group.get_first_member()
    node_name = first_control_plane.get_node_name()

    init_config = components.get_init_config(cluster, first_control_plane, init=True)
    config = components.get_kubeadm_config(cluster, init_config)

    utils.dump_file(cluster, config, 'init-config_%s.yaml' % node_name)

    log.debug("Uploading init config to initial control_plane...")
    first_control_plane.sudo("mkdir -p /etc/kubernetes")
    first_control_plane.put(io.StringIO(config), '/etc/kubernetes/init-config.yaml', sudo=True)

    # put control-plane patches
    components.create_kubeadm_patches_for_node(cluster, first_control_plane)

    # copy admission config to first control-plane
    first_control_plane.call(admission.copy_pss)

    # put audit-policy.yaml
    prepare_audit_policy(first_control_plane)

    log.debug("Initializing first control_plane...")
    result = first_control_plane.sudo(
        "kubeadm init"
        " --upload-certs"
        " --config=/etc/kubernetes/init-config.yaml"
        " --ignore-preflight-errors='" + cluster.inventory['services']['kubeadm_flags']['ignorePreflightErrors'] + "'"
        " --v=5",
        hide=False, pty=True)

    copy_admin_config(log, first_control_plane)

    kubeconfig_filepath = fetch_admin_config(cluster)
    summary.schedule_report(cluster.context, summary.SummaryItem.KUBECONFIG, kubeconfig_filepath)

    # Remove default resolvConf from kubelet-config ConfigMap for debian OS family
    first_control_plane.call(components.patch_kubelet_configmap)

    # Preparing join_dict to init other nodes
    control_plane_lines = list(result.values())[0].stdout. \
                       split("You can now join any number of the control-plane")[1].splitlines()[2:5]
    worker_lines = list(result.values())[0].stdout. \
                       split("Then you can join any number of worker")[1].splitlines()[2:4]
    control_plane_join_command = " ".join([x.replace("\\", "").strip() for x in control_plane_lines])
    worker_join_command = " ".join([x.replace("\\", "").strip() for x in worker_lines])

    # TODO: Get rid of this code and use get_join_dict() method
    args = control_plane_join_command.split("--")
    join_dict = {}
    for arg in args:
        key_val = arg.split(" ")
        if len(key_val) > 1:
            join_dict[key_val[0].strip()] = key_val[1].strip()
    join_dict["worker_join_command"] = worker_join_command
    cluster.context["join_dict"] = join_dict

    components.wait_for_pods(first_control_plane)
    # refresh cluster installation status in cluster context
    is_cluster_installed(cluster)


def wait_uncordon(node: NodeGroup) -> None:
    cluster = node.cluster
    timeout_config = cluster.inventory['globals']['expect']['pods']['kubernetes']
    # This forces to use local API server and waits till it is up.
    with local_admin_config(node) as kubeconfig:
        node.wait_command_successful(f"kubectl --kubeconfig {kubeconfig} uncordon {node.get_node_name()} > /dev/null",
                                     hide=False, pty=True,
                                     timeout=timeout_config['timeout'],
                                     retries=timeout_config['retries'])


def wait_for_nodes(group: NodeGroup) -> None:
    cluster: KubernetesCluster = group.cluster
    log = cluster.log

    first_control_plane = cluster.nodes["control-plane"].get_first_member()
    node_names = group.get_nodes_names()

    wait_conditions = {
        "Ready": "True",
        "NetworkUnavailable": "False"
    }
    if len(node_names) > 1:
        status_cmd = "kubectl get nodes %s -o jsonpath='{.items[*].status.conditions[?(@.type==\"%s\")].status}{\"\\n\"}'"
    else:
        status_cmd = "kubectl get nodes %s -o jsonpath='{.status.conditions[?(@.type==\"%s\")].status}{\"\\n\"}'"

    timeout = int(cluster.inventory['globals']['nodes']['ready']['timeout'])
    retries = int(cluster.inventory['globals']['nodes']['ready']['retries'])
    log.debug("Waiting for new kubernetes nodes to become ready, %s retries every %s seconds" % (retries, timeout))
    while retries > 0:
        correct_conditions = 0
        for condition, cond_value in wait_conditions.items():
            result = first_control_plane.sudo(status_cmd % (" ".join(node_names), condition), warn=True, pty=True)
            node_result = result.get_simple_result()
            if node_result.failed:
                log.debug(f"kubectl exited with non-zero exit code. Haproxy or kube-apiserver are not yet started?")
                log.verbose(node_result)
                break
            for line in node_result.stdout.rstrip('\n').split('\n'):
                condition_results = line.split(" ")
                correct_values = [value for value in condition_results if value == cond_value]
                if len(correct_values) == len(node_names):
                    correct_conditions = correct_conditions + 1
                    log.debug(f"Condition {condition} is {cond_value} for all nodes.")
                    break
            else:
                log.debug(f"Condition {condition} is not met, retrying")
                break

        if correct_conditions == len(wait_conditions):
            log.debug("All nodes are ready!")
            return
        else:
            retries = retries - 1
            time.sleep(timeout)

    raise Exception(f"Nodes did not become ready in the expected time, {retries} retries every {timeout} seconds. "
                    "Try to increase node.ready.retries parameter in globals: "
                    "https://github.com/Netcracker/KubeMarine/blob/main/documentation/Installation.md#globals")


def init_workers(group: NodeGroup) -> None:
    if group.is_empty():
        return

    cluster: KubernetesCluster = group.cluster
    join_dict = cluster.context.get("join_dict", get_join_dict(group))

    join_config = components.get_init_config(cluster, group, init=False, join_dict=join_dict)
    config = yaml.dump(join_config)

    utils.dump_file(cluster, config, 'join-config-workers.yaml')

    group.sudo("mkdir -p /etc/kubernetes")
    group.put(io.StringIO(config), '/etc/kubernetes/join-config.yaml', sudo=True)

    # put control-plane patches
    for node in group.get_ordered_members_list():
        components.create_kubeadm_patches_for_node(cluster, node)

    cluster.log.debug('Joining workers...')

    for node in group.get_ordered_members_list():
        node.sudo(
            "kubeadm join --config=/etc/kubernetes/join-config.yaml"
            " --ignore-preflight-errors='" + cluster.inventory['services']['kubeadm_flags']['ignorePreflightErrors'] + "'"
            " --v=5",
            hide=False, pty=True)

        components.wait_for_pods(node)


def apply_labels(group: NodeGroup) -> RunnersGroupResult:
    cluster: KubernetesCluster = group.cluster
    log = cluster.log

    log.debug("Applying additional labels for nodes")
    # TODO: Add "--overwrite-labels" switch
    # TODO: Add labels validation after applying
    control_plane = cluster.nodes["control-plane"].get_first_member()
    with control_plane.new_executor() as exe:
        for node in group.get_ordered_members_configs_list():
            if "labels" not in node:
                log.verbose("No additional labels found for %s" % node['name'])
                continue
            log.verbose("Found additional labels for %s: %s" % (node['name'], node['labels']))
            for key, value in node["labels"].items():
                exe.group.sudo("kubectl label node %s %s=%s" % (node["name"], key, value))

    log.debug("Successfully applied additional labels")

    return control_plane.sudo("kubectl get nodes --show-labels")


def apply_taints(group: NodeGroup) -> RunnersGroupResult:
    cluster: KubernetesCluster = group.cluster
    log = cluster.log

    log.debug("Applying additional taints for nodes")
    control_plane = cluster.nodes["control-plane"].get_first_member()
    with control_plane.new_executor() as exe:
        for node in group.get_ordered_members_configs_list():
            if "taints" not in node:
                log.verbose("No additional taints found for %s" % node['name'])
                continue
            log.verbose("Found additional taints for %s: %s" % (node['name'], node['taints']))
            for taint in node["taints"]:
                exe.group.sudo("kubectl taint node %s %s" % (node["name"], taint))

    log.debug("Successfully applied additional taints")

    return control_plane.sudo(
        "kubectl get nodes -o=jsonpath="
        "'{range .items[*]}{\"node: \"}{.metadata.name}{\"\\ntaints: \"}{.spec.taints}{\"\\n\"}{end}'")


def is_cluster_installed(cluster: KubernetesCluster) -> bool:
    cluster.log.verbose('Searching for already installed cluster...')
    try:
        results = cluster.nodes['control-plane'].sudo(
            'kubectl cluster-info', pty=True, warn=True, timeout=15)
        for host, result in results.items():
            if 'is running at' in result.stdout:
                cluster.log.verbose('Detected running Kubernetes cluster on %s' % host)
                for line in result.stdout.split("\n"):
                    if 'Kubernetes control plane' in line:
                        cluster.context['controlplain_uri'] = line.split('at ')[1]
                return True
    except Exception as e:
        cluster.log.verbose(e)
    cluster.context['controlplain_uri'] = None
    cluster.log.verbose('Failed to detect any Kubernetes cluster')
    return False


def upgrade_first_control_plane(upgrade_group: NodeGroup, cluster: KubernetesCluster, **drain_kwargs: Any) -> None:
    version = cluster.inventory["services"]["kubeadm"]["kubernetesVersion"]
    first_control_plane = cluster.nodes['control-plane'].get_first_member()
    node_name = first_control_plane.get_node_name()

    if not upgrade_group.has_node(node_name):
        cluster.log.debug("First control-plane \"%s\" upgrade is not required" % node_name)
        return

    cluster.log.debug("Upgrading first control-plane \"%s\"" % node_name)

    # put control-plane patches
    components.create_kubeadm_patches_for_node(cluster, first_control_plane)

    flags = ("-f --certificate-renewal=true "
             f"--ignore-preflight-errors='{cluster.inventory['services']['kubeadm_flags']['ignorePreflightErrors']}' "
             f"--patches=/etc/kubernetes/patches")

    drain_cmd = prepare_drain_command(cluster, node_name, **drain_kwargs)
    first_control_plane.sudo(drain_cmd, hide=False, pty=True)

    upgrade_cri_if_required(first_control_plane)
    fix_flag_kubelet(first_control_plane)

    first_control_plane.sudo(
        f"sudo kubeadm upgrade apply {version} {flags} && "
        f"sudo kubectl uncordon {node_name} && "
        f"sudo systemctl restart kubelet", hide=False, pty=True)

    copy_admin_config(cluster.log, first_control_plane)

    expect_kubernetes_version(cluster, version, apply_filter=node_name)
    components.wait_for_pods(first_control_plane)
    exclude_node_from_upgrade_list(first_control_plane, node_name)


def upgrade_other_control_planes(upgrade_group: NodeGroup, cluster: KubernetesCluster, **drain_kwargs: Any) -> None:
    version = cluster.inventory["services"]["kubeadm"]["kubernetesVersion"]
    first_control_plane = cluster.nodes['control-plane'].get_first_member()

    for node in cluster.nodes['control-plane'].get_ordered_members_list():
        node_name = node.get_node_name()
        if node_name != first_control_plane.get_node_name():

            if not upgrade_group.has_node(node_name):
                cluster.log.debug("Control-plane \"%s\" upgrade is not required" % node_name)
                continue

            cluster.log.debug("Upgrading control-plane \"%s\"" % node_name)

            # put control-plane patches
            components.create_kubeadm_patches_for_node(cluster, node)

            drain_cmd = prepare_drain_command(cluster, node_name, **drain_kwargs)
            node.sudo(drain_cmd, hide=False, pty=True)

            upgrade_cri_if_required(node)
            fix_flag_kubelet(node)

            node.sudo(
                f"sudo kubeadm upgrade node --certificate-renewal=true --patches=/etc/kubernetes/patches && "
                f"sudo kubectl uncordon {node_name} && "
                f"sudo systemctl restart kubelet",
                hide=False, pty=True)

            expect_kubernetes_version(cluster, version, apply_filter=node_name)
            copy_admin_config(cluster.log, node)
            components.wait_for_pods(node)
            exclude_node_from_upgrade_list(first_control_plane, node_name)


def upgrade_workers(upgrade_group: NodeGroup, cluster: KubernetesCluster, **drain_kwargs: Any) -> None:
    version = cluster.inventory["services"]["kubeadm"]["kubernetesVersion"]
    first_control_plane = cluster.nodes['control-plane'].get_first_member()

    for node in cluster.nodes['worker'].exclude_group(cluster.nodes['control-plane'])\
            .get_ordered_members_list():
        node_name = node.get_node_name()

        if not upgrade_group.has_node(node_name):
            cluster.log.debug("Worker \"%s\" upgrade is not required" % node_name)
            continue

        cluster.log.debug("Upgrading worker \"%s\"" % node_name)

        # put control-plane patches
        components.create_kubeadm_patches_for_node(cluster, node)

        drain_cmd = prepare_drain_command(cluster, node_name, **drain_kwargs)
        first_control_plane.sudo(drain_cmd, hide=False, pty=True)

        upgrade_cri_if_required(node)
        fix_flag_kubelet(node)

        node.sudo(
            "kubeadm upgrade node --certificate-renewal=true --patches=/etc/kubernetes/patches && "
            "sudo systemctl restart kubelet", pty=True)

        first_control_plane.sudo("kubectl uncordon %s" % node_name, hide=False)

        expect_kubernetes_version(cluster, version, apply_filter=node_name)
        # workers do not have system pods to wait for their start
        exclude_node_from_upgrade_list(first_control_plane, node_name)

        components.wait_for_pods(node)


def prepare_drain_command(cluster: KubernetesCluster, node_name: str,
                          *,
                          disable_eviction: bool = False,
                          drain_timeout: int = None, grace_period: int = None) -> str:
    drain_globals = static.GLOBALS['nodes']['drain']
    if drain_timeout is None:
        drain_timeout = recalculate_proper_timeout(cluster, drain_globals['timeout'])

    if grace_period is None:
        grace_period = drain_globals['grace_period']

    drain_cmd = f"kubectl drain {node_name} --force --ignore-daemonsets --delete-emptydir-data " \
                f"--timeout={drain_timeout}s --grace-period={grace_period}"
    if disable_eviction:
        drain_cmd += " --disable-eviction=true"
    return drain_cmd


def upgrade_cri_if_required(group: NodeGroup) -> None:
    # currently it is invoked only for single node
    cluster: KubernetesCluster = group.cluster
    log = cluster.log

    if 'containerd' in cluster.context["upgrade"]["required"]['packages']:
        cri_packages = cluster.get_package_association_for_node(group.get_host(), 'containerd', 'package_name')

        log.debug(f"Installing {cri_packages} on node: {group.get_node_name()}")
        packages.install(group, include=cri_packages, pty=True)
        log.debug(f"Restarting all containers on node: {group.get_node_name()}")
        group.sudo("crictl rm -fa", warn=True)
    else:
        log.debug("'containerd' package upgrade is not required")

    # upgrade of sandbox_image is currently not supported for migrate_kubemarine
    if cluster.context["upgrade"]["required"].get('containerdConfig', False):
        containerd.configure_containerd(group)
    else:
        log.debug("'containerd' configuration upgrade is not required")


def verify_upgrade_versions(cluster: KubernetesCluster) -> None:
    first_control_plane = cluster.nodes['control-plane'].get_first_member()
    upgrade_version = get_kubernetes_version(cluster.inventory)

    k8s_nodes_group = cluster.make_group_from_roles(['control-plane', 'worker'])
    for node in k8s_nodes_group.get_ordered_members_list():
        cluster.log.debug(f"Verifying current k8s version for node {node.get_node_name()}")
        result = first_control_plane.sudo("kubectl get nodes "
                                          f"{node.get_node_name()}"
                                          " -o custom-columns='VERSION:.status.nodeInfo.kubeletVersion' "
                                          "| grep -vw ^VERSION ")
        curr_version = list(result.values())[0].stdout
        test_version_upgrade_possible(curr_version, upgrade_version, skip_equal=True)


def get_procedure_upgrade_version(cluster: KubernetesCluster) -> str:
    upgrade_version: str = cluster.procedure_inventory['upgrade_plan'][cluster.context["upgrade_step"]]
    return upgrade_version


def get_kubernetes_version(inventory: dict) -> str:
    kubernetes_version: str
    if inventory.get("services", {}).get("kubeadm", {}).get("kubernetesVersion") is not None:
        kubernetes_version = str(inventory['services']['kubeadm']['kubernetesVersion'])
    else:
        kubernetes_version = static.DEFAULTS['services']['kubeadm']['kubernetesVersion']

    return kubernetes_version


@enrichment(EnrichmentStage.FULL)
def verify_version(cluster: KubernetesCluster) -> None:
    version = get_kubernetes_version(cluster.inventory)
    verify_allowed_version(version)
    verify_supported_version(version, cluster.log)


def verify_allowed_version(version: str) -> str:
    allowed_versions = static.KUBERNETES_VERSIONS['compatibility_map'].keys()
    if version not in allowed_versions:
        raise errors.KME('KME0008',
                         version=version,
                         allowed_versions=', '.join(map(repr, allowed_versions)))

    return version


def verify_supported_version(target_version: str, logger: log.EnhancedLogger) -> None:
    minor_version = utils.minor_version(target_version)
    supported_versions = static.KUBERNETES_VERSIONS['kubernetes_versions']
    if not supported_versions.get(minor_version, {}).get("supported", False):
        logger.warning(f"Specified target Kubernetes version {target_version!r} - is not supported!")


def expect_kubernetes_version(cluster: KubernetesCluster, version: str,
                              timeout: int = None, retries: int = None,
                              node: NodeGroup = None, apply_filter: str = None) -> None:
    if timeout is None:
        timeout = cluster.globals['nodes']['expect']['kubernetes_version']['timeout']
    if retries is None:
        retries = cluster.globals['nodes']['expect']['kubernetes_version']['retries']

    cluster.log.debug("Expecting Kubernetes version %s" % version)
    cluster.log.debug("Max expectation time: %ss" % (timeout * retries))

    cluster.log.debug("Waiting for nodes...\n")

    if node is None:
        node = cluster.nodes['control-plane'].get_first_member()

    command = 'kubectl get nodes -o=wide'
    if apply_filter is not None:
        command += ' | grep -w %s' % apply_filter

    while retries > 0:
        result = node.sudo(command, warn=True, pty=True)
        stdout = list(result.values())[0].stdout
        cluster.log.verbose(stdout)
        nodes_version_correct = True
        for stdout_line in iter(stdout.splitlines()):
            if version not in stdout_line:
                nodes_version_correct = False
                cluster.log.verbose("Invalid version detected: %s\n" % stdout_line)

        if nodes_version_correct:
            cluster.log.debug("Nodes have correct Kubernetes version = %s" % version)
            cluster.log.debug(result)
            return
        else:
            retries -= 1
            cluster.log.debug("Some nodes have invalid Kubernetes version... (%ss left)" % (retries * timeout), result)
            time.sleep(timeout)

    raise Exception('In the expected time, the nodes did not receive correct Kubernetes version')


def is_version_upgrade_possible(old: str, new: str) -> bool:
    try:
        test_version_upgrade_possible(old, new)
        return True
    except Exception:
        return False


def test_version_upgrade_possible(old: str, new: str, skip_equal: bool = False) -> None:
    old = old.strip()
    new = new.strip()
    old_version_key = utils.version_key(old)
    new_version_key = utils.version_key(new)

    # test new is greater than old
    if old_version_key > new_version_key:
        raise Exception(ERROR_DOWNGRADE % (old, new))

    # test new is the same as old
    if old_version_key == new_version_key and not skip_equal:
        raise Exception(ERROR_SAME % (old, new))

    # test major step is not greater than 1
    if new_version_key[0] - old_version_key[0] > 1:
        raise Exception(ERROR_MAJOR_RANGE_EXCEEDED % (old, new))

    # test minor step is not greater than 1
    if new_version_key[1] - old_version_key[1] > 1:
        raise Exception(ERROR_MINOR_RANGE_EXCEEDED % (old, new))

    # test the target version is the latest supported patch version
    new_minor_version = utils.minor_version(new)
    latest_supported_patch_version = max(
        (v for v in static.KUBERNETES_VERSIONS['compatibility_map']
         if utils.minor_version(v) == new_minor_version),
        key=utils.version_key)

    if new != latest_supported_patch_version:
        raise Exception(ERROR_NOT_LATEST_PATCH % (new, latest_supported_patch_version))


def recalculate_proper_timeout(cluster: KubernetesCluster, timeout: int) -> int:
    try:
        amount_str = cluster.nodes['control-plane'].get_first_member().sudo('kubectl get pods -A | wc -l').get_simple_out()
        return timeout * int(amount_str)
    except Exception:
        return timeout * 10 * cluster.nodes['all'].nodes_amount()


def exclude_node_from_upgrade_list(first_control_plane: NodeGroup, node_name: str) -> None:
    first_control_plane.sudo('sed -i \'/%s/d\' /etc/kubernetes/nodes-k8s-versions.txt' % node_name, warn=True)


def autodetect_non_upgraded_nodes(cluster: KubernetesCluster, future_version: str) -> List[str]:
    first_control_plane = cluster.nodes['control-plane'].get_first_member()
    try:
        nodes_list_result = (first_control_plane.sudo('[ ! -f /etc/kubernetes/nodes-k8s-versions.txt ] || '
                                                     'sudo cat /etc/kubernetes/nodes-k8s-versions.txt')
                                .get_simple_out())
        if not nodes_list_result:
            cluster.log.debug('Cluster status file /etc/kubernetes/nodes-k8s-versions.txt unexist or empty, '
                              'the new file will be created.')
            nodes_list_result = first_control_plane.sudo('sudo kubectl get nodes --no-headers -o custom-columns=\''
                                              'VERSION:.status.nodeInfo.kubeletVersion,'
                                              'NAME:.metadata.name,'
                                              'STATUS:.status.conditions[-1].type\'').get_simple_result().stdout
            nodes_list_result = (f'# This file contains a cached list of nodes and versions '
                                   'required to continue the Kubernetes upgrade procedure if it fails. '
                                   'If all the nodes are completely updated or you manually fixed the '
                                   f'problem that occurred during the upgrade, you can delete it.\n{nodes_list_result}')
            first_control_plane.put(io.StringIO(nodes_list_result), '/etc/kubernetes/nodes-k8s-versions.txt',
                                    backup=False, sudo=True)
        cluster.log.verbose("Remote response with nodes description:\n%s" % nodes_list_result)
    except Exception as e:
        cluster.log.warning("Failed to detect cluster status before upgrade. All nodes will be scheduled for upgrade.")
        cluster.log.verbose(e)
        return cluster.nodes['all'].get_nodes_names()

    detected_nodes_lines = nodes_list_result.splitlines()

    if not detected_nodes_lines:
        raise Exception('Remote result did not returned any lines containing node info')

    upgrade_list = []
    for line in detected_nodes_lines:
        line = line.strip()

        # comes from nodes-k8s-versions.txt content as a comment symbol
        if line[0] == '#':
            continue
        version, node_name, status = list(filter(len, line.split(' ')))
        if version != future_version:
            cluster.log.verbose("Node \"%s\" has version \"%s\" and scheduled for upgrade." % (node_name, version))
            upgrade_list.append(node_name)
        elif status != 'Ready':
            cluster.log.verbose("Node \"%s\" is not ready and scheduled for upgrade." % node_name)
            upgrade_list.append(node_name)
        else:
            cluster.log.verbose("Node \"%s\" already upgraded." % node_name)

    return upgrade_list


def get_group_for_upgrade(cluster: KubernetesCluster) -> NodeGroup:
    upgrade_group: Optional[NodeGroup] = cluster.context.get('upgrade_group')
    if upgrade_group is not None:
        return upgrade_group

    version = cluster.inventory["services"]["kubeadm"]["kubernetesVersion"]
    if cluster.procedure_inventory.get('upgrade_nodes'):
        nodes_for_upgrade = []
        for node in cluster.procedure_inventory['upgrade_nodes']:
            if isinstance(node, str):
                node_name = node
            else:
                node_name = node['name']
            nodes_for_upgrade.append(node_name)
            cluster.log.verbose("Node \"%s\" manually scheduled for upgrade." % node_name)
            cluster.nodes['control-plane'].get_first_member().sudo('rm -f /etc/kubernetes/nodes-k8s-versions.txt', warn=True)
    else:
        nodes_for_upgrade = autodetect_non_upgraded_nodes(cluster, version)

    upgrade_group = cluster.make_group_from_nodes(nodes_for_upgrade)
    cluster.context['upgrade_group'] = upgrade_group

    return upgrade_group


def images_grouped_prepull(group: NodeGroup, group_size: int = None) -> RunnersGroupResult:
    """
    Prepull kubeadm images on group, separated on sub-groups with certain group size. Separation required to avoid high
    load on images repository server, when using large clusters.
    :param group: NodeGroup where prepull should be performed.
    :param group_size: integer number of nodes per group. Will be automatically used from procedure_yaml or globals, if not set.
    :return: String results from all nodes in presented group.
    """

    cluster: KubernetesCluster = group.cluster
    log = cluster.log

    if group_size is None:
        group_size = cluster.procedure_inventory.get('prepull_group_size')

    if group_size is None:
        log.verbose("Group size is not set in procedure inventory, a default one will be used")
        group_size = cluster.globals['prepull_group_size']

    nodes_amount = group.nodes_amount()

    # group_size should be greater than 0
    if nodes_amount != 0 and nodes_amount < group_size:
        group_size = nodes_amount

    groups_amount = math.ceil(nodes_amount / group_size)

    log.verbose("Nodes amount: %s\nGroup size: %s\nGroups amount: %s" % (nodes_amount, group_size, groups_amount))
    collector = CollectorCallback(cluster)
    with group.new_executor() as exe:
        nodes = exe.group.get_ordered_members_list()
        for group_i in range(groups_amount):
            log.verbose('Prepulling images for group #%s...' % group_i)
            # RemoteExecutor used for future cases, when some nodes will require another/additional actions for prepull
            for node_i in range(group_i*group_size, (group_i*group_size)+group_size):
                if node_i < nodes_amount:
                    images_prepull(nodes[node_i], collector=collector)

    return collector.result


def images_prepull(group: DeferredGroup, collector: CollectorCallback) -> Token:
    """
    Prepull kubeadm images on group.

    :param group: single-node NodeGroup where prepull should be performed.
    :param collector: CollectorCallback instance
    :return: NodeGroupResult from all nodes in presented group.
    """

    cluster: KubernetesCluster = group.cluster
    kubeadm_init = components.get_init_config(cluster, group, init=True)
    config = components.get_kubeadm_config(cluster, kubeadm_init)

    group.put(io.StringIO(config), '/etc/kubernetes/prepull-config.yaml', sudo=True)

    return group.sudo("kubeadm config images pull --config=/etc/kubernetes/prepull-config.yaml",
                      pty=True, callback=collector)


def schedule_running_nodes_report(cluster: KubernetesCluster) -> None:
    summary.schedule_delayed_report(cluster, exec_running_nodes_report)


def exec_running_nodes_report(cluster: KubernetesCluster) -> None:
    nodes_description = get_nodes_description(cluster)
    actual_roles = get_actual_roles(nodes_description)
    nodes_conditions = get_nodes_conditions(nodes_description)
    nodes_names = actual_roles.keys()
    for role in ('control-plane', 'worker'):
        members = 0
        ready = 0
        for name in nodes_names:
            if role in actual_roles[name]:
                members += 1
                conditions = nodes_conditions[name]
                if conditions.get('Ready', {}).get('status') == 'True' \
                        and conditions.get('NetworkUnavailable', {}).get('status') == 'False':
                    ready += 1

        property_ = summary.SummaryItem.CONTROL_PLANES if role == 'control-plane' else summary.SummaryItem.WORKERS
        value = f'{ready}/{members}'
        summary.schedule_report(cluster.context, property_, value)


def get_nodes_description_cmd() -> str:
    return 'kubectl get node -o yaml'


def get_nodes_description(cluster: KubernetesCluster) -> dict:
    cmd = get_nodes_description_cmd()
    result = cluster.nodes['control-plane'].get_any_member().sudo(cmd)
    cluster.log.verbose(result)
    data: dict = yaml.safe_load(list(result.values())[0].stdout)
    return data


def get_actual_roles(nodes_description: dict) -> Dict[str, List[str]]:
    result: Dict[str, List[str]] = {}
    for node_description in nodes_description['items']:
        node_name = node_description['metadata']['name']
        labels = node_description['metadata']['labels']
        result[node_name] = []
        if 'node-role.kubernetes.io/control-plane' in labels:
            result[node_name].append('control-plane')
        if 'node-role.kubernetes.io/worker' in labels:
            result[node_name].append('worker')

    return result


def get_nodes_conditions(nodes_description: dict) -> Dict[str, Dict[str, dict]]:
    result = {}
    for node_description in nodes_description['items']:
        node_name = node_description['metadata']['name']
        conditions_by_type: Dict[str, dict] = {}
        result[node_name] = conditions_by_type
        for condition in node_description['status']['conditions']:
            conditions_by_type[condition['type']] = condition

    return result


def fix_flag_kubelet(group: NodeGroup) -> bool:
    kubeadm_flags_file = "/var/lib/kubelet/kubeadm-flags.env"
    cluster = group.cluster
    sandbox_image = containerd.get_sandbox_image(cluster.inventory['services']['cri'])
    infra_image_flag = f"--pod-infra-container-image={sandbox_image}"
    container_runtime_flag = '--container-runtime=remote'

    collector = CollectorCallback(cluster)
    with group.new_executor() as exe:
        for node in exe.group.get_ordered_members_list():
            node.sudo(f"cat {kubeadm_flags_file}", callback=collector)

    with group.new_executor() as exe:
        for node in exe.group.get_ordered_members_list():
            kubeadm_flags = collector.result[node.get_host()].stdout
            updated_kubeadm_flags = kubeadm_flags
            if components.is_container_runtime_not_configurable(cluster):
                # remove the deprecated kubelet flag for versions starting from 1.27.0
                updated_kubeadm_flags = updated_kubeadm_flags.replace(container_runtime_flag, '')

            if infra_image_flag not in updated_kubeadm_flags:
                # patch --pod-infra-container-image with target sandbox_image
                updated_kubeadm_flags = config_changer(updated_kubeadm_flags, infra_image_flag)

            if kubeadm_flags != updated_kubeadm_flags:
                cluster.log.debug(f"Patching {kubeadm_flags_file} on {node.get_node_name()} node...")
                node.put(io.StringIO(updated_kubeadm_flags), kubeadm_flags_file, backup=True, sudo=True)

    # If file is changed on at least one node, last results will be not empty
    return len(exe.get_last_results()) > 0


def config_changer(config: str, word: str) -> str:
    equal_pos = word.find("=") + 1
    param_begin_pos = config.find(word[:equal_pos])
    if param_begin_pos != -1:
        param_end_pos = config[param_begin_pos:].find(" ")
        if param_end_pos == -1:
            return config[:param_begin_pos] + word + "\""
        return config[:param_begin_pos] + word + config[param_end_pos + param_begin_pos:]
    else:
        param_end_pos = config.rfind("\"")
        return config[:param_end_pos] + " " + word[:] + "\""


def prepare_audit_policy(group: NodeGroup) -> None:
    """
    Prepare audit-policy.yaml and all necessary directories.
    """
    cluster: KubernetesCluster = group.cluster
    api_server_extra_args = cluster.inventory['services']['kubeadm']['apiServer']['extraArgs']
    audit_log_dir = os.path.dirname(api_server_extra_args['audit-log-path'])
    audit_file_name = api_server_extra_args['audit-policy-file']
    audit_policy_dir = os.path.dirname(audit_file_name)
    group.sudo(f"mkdir -p {audit_log_dir} && sudo mkdir -p {audit_policy_dir}")

    cluster.log.debug("Configure audit cluster policy")
    policy_config = cluster.inventory['services']['audit']['cluster_policy']
    policy_config_file = yaml.dump(policy_config)
    utils.dump_file(cluster, policy_config_file, 'audit-policy.yaml')
    # upload rules on cluster
    group.put(io.StringIO(policy_config_file), audit_file_name, sudo=True, backup=True)



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
from typing import Optional

import yaml

from kubemarine.core import utils, summary
from kubemarine.core.cluster import KubernetesCluster, EnrichmentStage, enrichment


@enrichment(EnrichmentStage.FULL)
def enrich_inventory(cluster: KubernetesCluster) -> None:
    inventory = cluster.inventory
    rbac = inventory['rbac']
    if not rbac.get("accounts"):
        return

    for i, account in enumerate(rbac["accounts"]):
        if account['configs'][0]['metadata'].get('name') is None:
            rbac["accounts"][i]['configs'][0]['metadata']['name'] = account['name']
        if account['configs'][0]['metadata'].get('namespace') is None:
            rbac["accounts"][i]['configs'][0]['metadata']['namespace'] = account['namespace']

        if account['configs'][1]['metadata'].get('name') is None:
            rbac["accounts"][i]['configs'][1]['metadata']['name'] = account['name']
        if account['configs'][1]['roleRef'].get('name') is None:
            rbac["accounts"][i]['configs'][1]['roleRef']['name'] = account['role']

        if account['configs'][1]['subjects'][0].get('name') is None:
            rbac["accounts"][i]['configs'][1]['subjects'][0]['name'] = account['name']
        if account['configs'][1]['subjects'][0].get('namespace') is None:
            rbac["accounts"][i]['configs'][1]['subjects'][0]['namespace'] = account['namespace']

        if account['configs'][2]['metadata'].get('name') is None:
            rbac["accounts"][i]['configs'][2]['metadata']['annotations']['kubernetes.io/service-account.name'] \
                = account['name']
            rbac["accounts"][i]['configs'][2]['metadata']['name'] = f"{account['name']}-token"
            rbac["accounts"][i]['configs'][0]['secrets'].append({})
            rbac["accounts"][i]['configs'][0]['secrets'][0]['name'] = f"{account['name']}-token"
        if account['configs'][2]['metadata'].get('namespace') is None:
            rbac["accounts"][i]['configs'][2]['metadata']['namespace'] = account['namespace']


def install(cluster: KubernetesCluster) -> None:
    rbac = cluster.inventory['rbac']
    if not rbac.get("accounts"):
        cluster.log.debug("No accounts specified to install, skipping...")
        return

    tokens = []
    for account in rbac["accounts"]:
        cluster.log.debug('Creating cluster account:')
        cluster.log.debug('\tName: %s\n\tRole: %s\n\tNameSpace: %s' % (account['name'], account['role'], account['namespace']))

        dump = ''
        for config in account['configs']:
            dump += '---\n'+yaml.dump(config, default_flow_style=False)

        filename = 'account_%s_%s_%s.yaml' % (account['name'], account['role'], account['namespace'])
        destination_path = '/etc/kubernetes/%s' % filename

        utils.dump_file(cluster, dump, filename)

        cluster.log.debug("Uploading template...")
        cluster.log.debug("\tDestination: %s" % destination_path)
        cluster.nodes['control-plane'].put(io.StringIO(dump), destination_path, sudo=True)

        cluster.log.debug("Applying yaml...")
        cluster.nodes['control-plane'].get_first_member().sudo('kubectl apply -f %s' % destination_path, hide=False)

        cluster.log.debug('Loading token...')
        load_tokens_cmd = 'kubectl -n %s get secret ' \
                          '$(sudo kubectl -n %s get sa %s -o \'jsonpath={.secrets[0].name}\') -o \'jsonpath={.data.token}\'' \
                          '| sudo base64 -d' % (account['namespace'], account['namespace'], account['name'])

        token: Optional[str] = None
        retries = cluster.globals['accounts']['retries']
        # Token creation is not synchronous, therefore retries are necessary
        while retries > 0:
            result = cluster.nodes['control-plane'].get_first_member().sudo(load_tokens_cmd)
            token = list(result.values())[0].stdout
            if not token:
                retries -= 1
            else:
                break
        if not token:
            raise Exception(f"The token loading for {account['name']} 'ServiceAccount' failed")

        tokens.append({
            'name': account['name'],
            'role': account['role'],
            'namespace': account['namespace'],
            'token': token,
        })

    cluster.log.debug('\nSaving tokens...')
    token_filename = utils.get_external_resource_path('account-tokens.yaml')
    utils.dump_file(cluster.context, yaml.dump(tokens), token_filename, dump_location=False)
    cluster.log.debug('Tokens saved to %s' % token_filename)

    summary.schedule_report(cluster.context, summary.SummaryItem.ACCOUNT_TOKENS, token_filename)



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module works with audit on remote nodes.
Using this module you can install, enable audit and configure audit rules.
"""

import io
import shlex
from typing import Optional, Tuple, List, Set

from kubemarine import system, packages
from kubemarine.core import utils
from kubemarine.core.cluster import KubernetesCluster, EnrichmentStage, enrichment
from kubemarine.core.group import NodeGroup, RunnersGroupResult, CollectorCallback


@enrichment(EnrichmentStage.FULL)
def verify_inventory(cluster: KubernetesCluster) -> None:
    for node in cluster.make_group_from_roles(['control-plane', 'worker']).get_accessible_nodes().get_ordered_members_list():
        if node.get_nodes_os() in ('unknown', 'unsupported'):
            continue
        host = node.get_host()
        package_name = cluster.get_package_association_for_node(host, 'audit', 'package_name')
        if isinstance(package_name, str):
            package_name = [package_name]

        if len(package_name) != 1:
            os_family = cluster.get_os_family_for_node(host)
            raise Exception(f'Audit has multiple associated packages {package_name} for OS {os_family!r} '
                            f'that is currently not supported')


def install(group: NodeGroup) -> Optional[RunnersGroupResult]:
    """
    Automatically installs and enables the audit service for the specified nodes
    :param group: Nodes group on which audit installation should be performed
    :return: String with installation output from nodes or None, when audit installation was skipped
    """
    cluster: KubernetesCluster = group.cluster
    log = cluster.log

    log.verbose('Searching for already installed auditd package...')

    # Reduce nodes amount for installation
    hosts_to_packages = packages.get_association_hosts_to_packages(group, cluster.inventory, 'audit')

    not_installed_hosts = []
    audit_installed_results = packages.detect_installed_packages_version_hosts(cluster, hosts_to_packages)
    for detected_audit_versions in audit_installed_results.values():
        for detected_version, hosts in detected_audit_versions.items():
            log.verbose(f'{detected_version}: {hosts}')
            if 'not installed' in detected_version:
                not_installed_hosts.extend(hosts)

    if not not_installed_hosts:
        log.debug('Auditd is already installed on all nodes')
        return None
    else:
        log.debug(f'Auditd package is not installed on {not_installed_hosts}, installing...')

    collector = CollectorCallback(cluster)
    with cluster.make_group(not_installed_hosts).new_executor() as exe:
        for node in exe.group.get_ordered_members_list():
            package_name = cluster.get_package_association_for_node(node.get_host(), 'audit', 'package_name')
            packages.install(node, include=package_name, pty=True, callback=collector)

            service_name = cluster.get_package_association_for_node(node.get_host(), 'audit', 'service_name')
            system.enable_service(node, name=service_name, callback=collector)

    return collector.result


def make_config(cluster: KubernetesCluster) -> str:
    return " \n".join(cluster.inventory['services']['audit']['rules'])


def apply_audit_rules(group: NodeGroup) -> RunnersGroupResult:
    """
    Generates and applies audit rules to the group
    :param group: Nodes group, where audit service should be configured
    :return: Service restart result or nothing if audit rules are non exists, or restart is not required
    """
    cluster: KubernetesCluster = group.cluster
    log = cluster.log

    rules_valid, auditctl_results = audit_rules_valid(group)
    if rules_valid:
        log.debug('Auditd rules are already actual on all nodes')
        return auditctl_results

    log.debug('Applying audit rules...')
    rules_content = make_config(cluster)
    utils.dump_file(cluster, rules_content, 'audit.rules')

    collector = CollectorCallback(cluster)
    with group.new_executor() as exe:
        for node in exe.group.get_ordered_members_list():
            host = node.get_host()

            rules_config_location = cluster.get_package_association_for_node(host, 'audit', 'config_location')
            node.put(io.StringIO(rules_content), rules_config_location, sudo=True, backup=True)

            service_name = cluster.get_package_association_for_node(host, 'audit', 'service_name')
            node.sudo(f'service {service_name} restart', callback=collector)

    _, auditctl_results = audit_rules_valid(group, silent=True)
    return auditctl_results


def audit_rules_valid(group: NodeGroup, silent: bool = False) -> Tuple[bool, RunnersGroupResult]:
    cluster: KubernetesCluster = group.cluster
    logger = cluster.log

    rules_content = make_config(cluster)

    collector = CollectorCallback(cluster)
    with group.new_executor() as exe:
        for node in exe.group.get_ordered_members_list():
            executable = cluster.get_package_association_for_node(node.get_host(), 'audit', 'executable_name')
            node.sudo(f'{executable} -l', callback=collector)

    verify_results = collector.result
    rules_valid = True
    for host, result in verify_results.items():
        tokens: List[Set[str]] = [set(shlex.split(line)) for line in result.stdout.rstrip('\n').split('\n')]
        for rule in rules_content.split('\n'):
            if not any(set(shlex.split(rule)).issubset(token) for token in tokens):
                if not silent:
                    logger.debug(f'Audit rule {rule!r} is not found at {host}')
                rules_valid = False

    return rules_valid, verify_results



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import re
from typing import Optional, Dict, Any, Tuple, List, Callable, Union, Sequence, cast, Type

import yaml

from kubemarine.core.cluster import KubernetesCluster, EnrichmentStage, enrichment
from kubemarine import jinja, keepalived, haproxy, kubernetes, thirdparties
from kubemarine.core import utils, static, log, os
from kubemarine.core.proxytypes import Primitive, Index, Node
from kubemarine.core.yaml_merger import default_merger
from kubemarine.cri.containerd import contains_old_format_properties


supported_connection_defaults = {
    'node_defaults': 'nodes',
}

supported_defaults = {
    'rbac': {
        'account_defaults': 'accounts'
    },
}

connection_sections = [
    'nodes', 'node_defaults', 'gateway_nodes',
    'procedure_history',  # Required to preserve inventory
    'values', 'cluster_name'  # May be referred to in jinja expressions
]

ssh_access_default_properties = [
    'keyfile', 'password', 'username', 'connection_port', 'connection_timeout', 'gateway',
    'boot'  # Participates in workarounds and reboot. This dependency can potentially be avoided
]

ssh_access_node_properties = ssh_access_default_properties + [
    'address', 'internal_address', 'connect_to',
    'name',  # In context of connections, the name is used only in logging. This dependency can potentially be avoided.
    'roles'  # Required to distinguish control planes to preserve inventory.
]


invalid_node_name_regex = re.compile("[^a-z-.\\d]", re.M)


@enrichment(EnrichmentStage.LIGHT, procedures=['add_node'])
def add_node_enrich_roles(cluster: KubernetesCluster) -> None:
    # At LIGHT stage, we should mark the nodes with `add_node` service role to find them after compilation.
    # This is for optimization to avoid two-staged enrichment even for connections only.
    for node in cluster.procedure_inventory['nodes']:
        # deepcopy is necessary, otherwise role append will happen in procedure_inventory too
        node = utils.deepcopy_yaml(node)
        node["roles"].append("add_node")
        cluster.inventory["nodes"].append(node)


@enrichment(EnrichmentStage.LIGHT, procedures=['remove_node'])
def remove_node_enrich_roles(cluster: KubernetesCluster) -> None:
    # At LIGHT stage, we should mark the nodes with `remove_node` service role to find them after compilation.
    # This is for optimization to avoid two-staged enrichment even for connections only.
    node_names_to_remove = [node['name'] for node in cluster.procedure_inventory["nodes"]]
    for node_remove in node_names_to_remove:
        for node in cluster.inventory['nodes']:
            # Inventory is not compiled at this step.
            # Expecting that the names are not jinja, or the same jinja expressions.
            if node['name'] == node_remove:
                node['roles'].append('remove_node')
                break
        else:
            raise Exception(f"Failed to find node to remove {node_remove} among existing nodes")


@enrichment(EnrichmentStage.PROCEDURE, procedures=['add_node'])
def enrich_add_nodes(cluster: KubernetesCluster) -> None:
    # Unlike at LIGHT stage (add_node_enrich_roles), at PROCEDURE stage we add the nodes without service roles.
    for new_node in cluster.procedure_inventory["nodes"]:
        node = utils.deepcopy_yaml(new_node)
        cluster.inventory["nodes"].append(node)


@enrichment(EnrichmentStage.PROCEDURE, procedures=['remove_node'])
def enrich_remove_nodes(cluster: KubernetesCluster) -> None:
    # Unlike at LIGHT stage (remove_node_enrich_roles), at PROCEDURE stage we remove the nodes completely.
    node_names_to_remove = [node['name'] for node in cluster.procedure_inventory["nodes"]]
    for node_remove in node_names_to_remove:
        for i, node in enumerate(cluster.inventory['nodes']):
            if node['name'] == node_remove:
                del cluster.inventory['nodes'][i]
                break


@enrichment(EnrichmentStage.LIGHT, procedures=['add_node', 'remove_node'])
def remove_service_roles(cluster: KubernetesCluster) -> None:
    for node in cluster.inventory['nodes']:
        roles = node['roles']
        for role in ('add_node', 'remove_node'):
            if role in roles:
                roles.remove(role)


@enrichment(EnrichmentStage.ALL)
def apply_connection_defaults(cluster: KubernetesCluster) -> None:
    recursive_apply_defaults(supported_connection_defaults, cluster.inventory)


@enrichment(EnrichmentStage.FULL)
def apply_defaults(cluster: KubernetesCluster) -> None:
    recursive_apply_defaults(supported_defaults, cluster.inventory)


@enrichment(EnrichmentStage.ALL)
def calculate_connect_to(cluster: KubernetesCluster) -> None:
    for node in cluster.inventory["nodes"]:
        address = cluster.get_access_address_from_node(node)
        # we definitely know how to connect
        node["connect_to"] = address


@enrichment(EnrichmentStage.ALL)
def calculate_nodegroups(cluster: KubernetesCluster) -> None:
    for nodes, skip_role in (
            (cluster.previous_nodes, 'add_node'),
            (cluster.nodes, 'remove_node'),
    ):
        if nodes:
            # Since this is the first enrichment procedure that fills the nodes,
            # having non-empty nodes means they are externally provided (e.g. at PROCEDURE stage).
            continue

        ips: Dict[str, List[str]] = {
            "all": []
        }

        for node in cluster.inventory["nodes"]:
            if skip_role in node['roles']:
                continue

            address = node['connect_to']
            ips['all'].append(address)
            for role in node['roles']:
                if role not in ('remove_node', 'add_node'):
                    ips.setdefault(role, []).append(address)

        for role, hosts in ips.items():
            nodes[role] = cluster.make_group(hosts)


@enrichment(EnrichmentStage.FULL)
def apply_registry(cluster: KubernetesCluster) -> None:
    inventory = cluster.inventory

    if not inventory.get('registry'):
        cluster.log.verbose('Unified registry is not used')
        return

    thirdparties_address = None
    containerd_endpoints = None
    protocol = None

    # registry contains either 'endpoints' or 'address' that is validated by JSON schema.
    if inventory['registry'].get('endpoints'):
        registry_mirror_address, containerd_endpoints, thirdparties_address = apply_registry_endpoints(inventory)
    else:
        if inventory['registry'].get('docker_port'):
            registry_mirror_address = "%s:%s" % (inventory['registry']['address'], inventory['registry']['docker_port'])
        else:
            registry_mirror_address = inventory['registry']['address']

        protocol = 'http'
        if inventory['registry'].get('ssl', False):
            protocol = 'https'

        if inventory['registry'].get('webserver', False):
            thirdparties_address = f"{protocol}://{inventory['registry']['address']}"

    # Patch kubeadm imageRepository and plugin_defaults registry
    if cluster.raw_inventory.get('services', {}).get('kubeadm', {}).get('imageRepository') is None:
        inventory['services']['kubeadm']["imageRepository"] = registry_mirror_address
    inventory['plugin_defaults']['installation'].setdefault('registry', registry_mirror_address)

    if not containerd_endpoints:
        containerd_endpoints = ["%s://%s" % (protocol, registry_mirror_address)]

    old_format_result, _ = contains_old_format_properties(inventory)
    if old_format_result:
        # Add registry info in old format
        registry_section = f'plugins."io.containerd.grpc.v1.cri".registry.mirrors."{registry_mirror_address}"'
        containerd_config = inventory['services']['cri']['containerdConfig']
        if not containerd_config.get(registry_section):
            containerd_config[registry_section] = {
                'endpoint': containerd_endpoints
            }
    else:
        # Add registry info in new format
        old_registry_config = inventory['services']['cri'].get('containerdRegistriesConfig', {})
        inventory['services']['cri']['containerdRegistriesConfig'] = {registry_mirror_address: {
            f'host."{endpoint}"': {'capabilities': ['pull', 'resolve']}
            for endpoint in containerd_endpoints
        }}
        default_merger.merge(inventory['services']['cri']['containerdRegistriesConfig'], old_registry_config)

    if thirdparties_address:
        for destination, config in inventory['services']['thirdparties'].items():
            if not thirdparties.is_default_thirdparty(destination) or isinstance(config, str) or 'source' in config:
                continue

            source, sha1 = thirdparties.get_default_thirdparty_identity(cluster.inventory, destination, in_public=False)
            source = source.format(registry=thirdparties_address)
            config['source'] = source
            if 'sha1' not in config:
                config['sha1'] = sha1


def apply_registry_endpoints(inventory: dict) -> Tuple[str, List[str], Optional[str]]:

    if not inventory['registry'].get('mirror_registry'):
        inventory['registry']['mirror_registry'] = 'registry.cluster.local'

    registry_mirror_address = inventory['registry']['mirror_registry']

    # todo Currently registry.endpoints is used only for containerd registry mirrors, but it can be provided explicitly.
    #  Probably we could make endpoints optional in this case.
    containerd_endpoints = inventory['registry']['endpoints']
    thirdparties_address = inventory['registry'].get('thirdparties')

    return registry_mirror_address, containerd_endpoints, thirdparties_address


@enrichment(EnrichmentStage.FULL)
def append_controlplain(cluster: KubernetesCluster) -> None:
    _append_controlplain(cluster.inventory, cluster.log)


def _append_controlplain(inventory: dict, logger: log.EnhancedLogger) -> None:

    if inventory.get('control_plain', {}).get('internal') and inventory.get('control_plain', {}).get('external'):
        logger.verbose('Control plains are set manually, nothing to detect.')

    logger.verbose('Detecting control plains...')

    # calculate controlplain ips
    internal_address: Optional[str] = None
    internal_address_source: Optional[str] = None
    external_address: Optional[str] = None
    external_address_source: Optional[str] = None

    balancer_names = keepalived.get_all_balancer_names(inventory)
    # vrrp_ip section is not enriched yet
    # If no VRRP IPs or no balancers are configured, Keepalived is not enabled.
    if inventory.get('vrrp_ips') and balancer_names:
        for i, item in enumerate(inventory['vrrp_ips']):
            if isinstance(item, str):
                if internal_address is None:
                    internal_address = item
                    internal_address_source = 'vrrp_ip[%s]' % i
            else:
                if haproxy.is_vrrp_not_bind(item):
                    continue
                final_hosts = item.get('hosts', balancer_names)
                # There is a small gap here.
                # The check is invoked when inventory is not yet compiled, so checking names for equality is not fair.
                if not any((host['name'] if isinstance(host, dict) else host) in balancer_names
                           for host in final_hosts):
                    continue
                if internal_address is None or item.get('control_endpoint', False):
                    internal_address = item['ip']
                    internal_address_source = 'vrrp_ip[%s]' % i
                if item.get('floating_ip') and (external_address is None or item.get('control_endpoint', False)):
                    external_address = item['floating_ip']
                    external_address_source = 'vrrp_ip[%s]' % i

    if internal_address is not None and external_address is None:
        logger.warning('VRRP_IPs has an internal address, but do not have an external one. '
                       'Your configuration may be incorrect. Trying to handle this problem automatically...')

    if internal_address is None or external_address is None:
        for role in ['balancer', 'control-plane']:
            # nodes are not compiled to groups yet
            for node in inventory['nodes']:
                if role in node['roles']:
                    if internal_address is None or node.get('control_endpoint', False):
                        internal_address = node['internal_address']
                        internal_address_source = f"{role} \"{node['name']}\""
                    if node.get('address') and (external_address is None or node.get('control_endpoint', False)):
                        external_address = node['address']
                        external_address_source = f"{role} \"{node['name']}\""

    if external_address is None:
        logger.warning('Failed to detect external control plain. Something may work incorrect!')
        external_address = internal_address

    logger.debug('Control plains:\n   Internal: %s (%s)\n   External: %s (%s)'
                 % (internal_address, internal_address_source, external_address, external_address_source))

    # apply controlplain ips
    if not inventory.get('control_plain'):
        inventory['control_plain'] = {}

    if not inventory['control_plain'].get('internal'):
        inventory['control_plain']['internal'] = internal_address

    if not inventory['control_plain'].get('external'):
        inventory['control_plain']['external'] = external_address


def recursive_apply_defaults(defaults: dict, section: dict) -> None:
    for key, value in defaults.items():
        if isinstance(value, dict):
            if section.get(key) is not None and section[key]:
                recursive_apply_defaults(value, section[key])
        # check if target section exists and not empty
        elif section.get(value) is not None:
            for i, custom_value in enumerate(section[value]):
                # copy defaults as new dict, to avoid problems with memory links
                default_value = utils.deepcopy_yaml(section[key])

                # update defaults with custom-defined node configs
                # TODO: Use deepmerge instead of update
                default_value.update(custom_value)

                # replace old node config with merged one
                section[value][i] = default_value


@enrichment(EnrichmentStage.ALL)
def calculate_node_names(cluster: KubernetesCluster) -> None:
    roles_iterators: Dict[str, int] = {}
    for node in cluster.inventory['nodes']:
        # 'master' role is not deleted because calculate_node_names() can be run over old inventory,
        # that may still have the old role (LIGHT enrichment).
        for role_name in ['control-plane', 'master', 'worker', 'balancer']:
            if role_name in node['roles']:
                # The idea is this:
                # If the name is already specified, we must skip this node,
                # however, we must consider that we already have a node of this type
                # and increase this type iterator
                # As a result, we get such an algorithm. For example, with the following inventory:
                #
                # - name: k8s-control-plane-1, roles: ['control-plane']
                # - roles: ['control-plane']
                # - name: k8s-control-plane-3, roles: ['control-plane']
                #
                # We should get the following calculation result:
                #
                # - name: k8s-control-plane-1, roles: ['control-plane']
                # - name: control-plane-2, roles: ['control-plane']
                # - name: k8s-control-plane-3, roles: ['control-plane']
                #
                role_i = roles_iterators.get(role_name, 1)
                roles_iterators[role_name] = role_i + 1
                if node.get('name') is None:
                    if role_name == 'master':
                        role_name = 'control-plane'
                    new_name = '%s-%s' % (role_name, role_i)
                    cluster.log.debug(f"Assigning name {new_name} to node {cluster.get_access_address_from_node(node)}")
                    node['name'] = new_name


@enrichment(EnrichmentStage.ALL)
def verify_nodes(cluster: KubernetesCluster) -> None:
    known_names = []
    known_hosts = []
    known_internal_addresses = []
    for node in cluster.inventory['nodes']:
        node_name = node['name']
        if node_name in known_names:
            raise Exception('Node name %s is duplicated in configfile' % node_name)
        if re.findall(invalid_node_name_regex, node_name):
            raise Exception('Node name \"%s\" contains invalid characters. A DNS-1123 subdomain must consist of lower '
                            'case alphanumeric characters, \'-\' or \'.\'' % node_name)
        known_names.append(node_name)

        host = node['connect_to']
        if host in known_hosts:
            raise Exception('Access address %s is duplicated in configfile' % host)
        known_hosts.append(host)

        internal_address = node['internal_address']
        if internal_address in known_internal_addresses:
            raise Exception('Internal address %s is duplicated in configfile' % internal_address)
        known_internal_addresses.append(internal_address)

    known_gateway_node_names = []
    for gateway_node in cluster.inventory.get('gateway_nodes', []):
        node_name = gateway_node['name']
        if node_name in known_gateway_node_names:
            raise Exception(f"Gateway node name {node_name} is duplicated in configfile")

        known_gateway_node_names.append(node_name)


def restrict_connection_sections(inventory: dict) -> dict:
    """
    Returns shallow copy of the inventory with only those sections
    that participate in the enrichment of connections.
    """
    inventory = utils.subdict_yaml(inventory, connection_sections)
    node_defaults = inventory.get('node_defaults', {})
    if node_defaults:
        inventory['node_defaults'] = utils.subdict_yaml(node_defaults, ssh_access_default_properties)

    nodes = inventory.get('nodes', [])
    if nodes:
        inventory['nodes'] = nodes = list(nodes)
        for i, node in enumerate(nodes):
            nodes[i] = utils.subdict_yaml(node, ssh_access_node_properties)

    return inventory


@enrichment(EnrichmentStage.LIGHT)
def restrict_connections(cluster: KubernetesCluster) -> dict:
    return restrict_connection_sections(cluster.inventory)


@enrichment(EnrichmentStage.LIGHT)
def merge_connection_defaults(cluster: KubernetesCluster) -> dict:
    connection_defaults = restrict_connection_sections(static.DEFAULTS)
    return _merge_inventory(cluster, connection_defaults)


@enrichment(EnrichmentStage.FULL)
def merge_defaults(cluster: KubernetesCluster) -> dict:
    return _merge_inventory(cluster, static.DEFAULTS)


def _merge_inventory(cluster: KubernetesCluster, base: dict) -> dict:
    base_inventory = utils.deepcopy_yaml(base)
    inventory: dict = default_merger.merge(base_inventory, cluster.inventory)
    return inventory


@enrichment(EnrichmentStage.LIGHT)
def compile_connections(cluster: KubernetesCluster) -> None:
    return _compile_inventory(cluster, light=True)


@enrichment(EnrichmentStage.FULL)
def compile_inventory(cluster: KubernetesCluster) -> None:
    return _compile_inventory(cluster, light=False)


def _compile_inventory(cluster: KubernetesCluster, *, light: bool) -> None:
    inventory = cluster.inventory

    extra: Dict[str, Any] = {'env': os.Environ()}
    if not light:
        extra['globals'] = static.GLOBALS

    if light:
        # Management of primitive values is currently not necessary for LIGHT stage.
        env = Environment(cluster.log, inventory, recursive_compile=True, recursive_extra=extra)
        jinja.compile_node(inventory, [], env)
    else:
        primitives_config = _get_primitive_values_registry()
        env = Environment(cluster.log, inventory, recursive_compile=True, recursive_extra=extra,
                          primitives_config=primitives_config)
        compile_node_with_primitives(inventory, [], env, primitives_config)

    remove_empty_items(inventory)

    dump_inventory(cluster, cluster.context, "cluster_precompiled.yaml")


def dump_inventory(cluster: KubernetesCluster, context: dict, filename: str) -> None:
    if not utils.is_dump_allowed(context, filename):
        return

    data = yaml.dump(cluster.inventory)
    utils.dump_file(context, data, filename)


PrimitivesConfig = List[Tuple[List[str], Callable[[Any], Any]]]


def compile_node_with_primitives(struct: Union[list, dict],
                                 path: List[Union[str, int]],
                                 env: jinja.Environment,
                                 primitives_config: PrimitivesConfig) -> Union[list, dict]:
    if isinstance(struct, list):
        for i, v in enumerate(struct):
            struct[i] = compile_object_with_primitives(v, path, i, env, primitives_config)
    else:
        for k, v in struct.items():
            struct[k] = compile_object_with_primitives(v, path, k, env, primitives_config)

    return struct


def compile_object_with_primitives(struct: Union[Primitive, list, dict],
                                   path: List[Index], index: Index,
                                   env: jinja.Environment,
                                   primitives_config: PrimitivesConfig) -> Union[Primitive, list, dict]:
    depth = len(path)
    primitives_config = choose_nested_primitives_config(primitives_config, depth, index)

    path.append(index)
    if isinstance(struct, (list, dict)):
        if primitives_config:
            struct = compile_node_with_primitives(struct, path, env, primitives_config)
        else:
            struct = jinja.compile_node(struct, path, env)
    else:
        if isinstance(struct, str) and jinja.is_template(struct):
            struct = env.compile_string(struct, jinja.Path(path))

        struct = convert_primitive(struct, path, primitives_config)

    path.pop()
    return struct


def remove_empty_items(struct: Any) -> Any:
    if isinstance(struct, list):
        new_struct = []
        for v in struct:
            v = remove_empty_items(v)
            # delete empty list entries, which can appear after jinja compilation
            if v != '':
                new_struct.append(v)
        struct = new_struct
    elif isinstance(struct, dict):
        for k, v in struct.items():
            struct[k] = remove_empty_items(v)

    return struct


def escape_jinja_characters_for_inventory(cluster: KubernetesCluster, obj: Any) -> Any:
    if isinstance(obj, dict):
        for key, value in obj.items():
            obj[key] = escape_jinja_characters_for_inventory(cluster, value)
    elif isinstance(obj, list):
        for key, value in enumerate(obj):
            obj[key] = escape_jinja_characters_for_inventory(cluster, value)
    elif isinstance(obj, str):
        obj = _escape_jinja_character(obj)
    return obj


def _escape_jinja_character(value: str) -> str:
    if jinja.is_template(value):
        value = '{{ %s }}' % (json.JSONEncoder().encode(value),)

    return value


def _get_primitive_values_registry() -> PrimitivesConfig:
    return [
        (['services', 'cri', 'containerdConfig',
          'plugins."io.containerd.grpc.v1.cri".containerd.runtimes.runc.options',
          'SystemdCgroup'], utils.strtobool),
        (['services', 'modprobe', '*', '*', 'install'], utils.strtobool),
        # kernel parameters are actually not always represented as integers
        (['services', 'sysctl', '*', 'value'], utils.strtoint),
        (['services', 'sysctl', '*', 'install'], utils.strtobool),
        # kernel parameters are actually not always represented as integers
        (['patches', '*', 'services', 'sysctl', '*', 'value'], utils.strtoint),
        (['patches', '*', 'services', 'sysctl', '*', 'install'], utils.strtobool),
        (['plugins', '*', 'install'], utils.strtobool),
        (['plugins', 'calico', 'typha', 'enabled'], utils.strtobool),
        (['plugins', 'calico', 'typha', 'replicas'], utils.strtoint),
        (['plugins', 'nginx-ingress-controller', 'ports', '*', 'hostPort'], utils.strtoint),
    ]


def choose_nested_primitives_config(primitives_config: PrimitivesConfig, depth: int, index: Index) -> PrimitivesConfig:
    nested_config = []
    for search in primitives_config:
        search_path = search[0]
        if depth == len(search_path):
            continue

        section = search_path[depth]
        if section in ('*', index):
            nested_config.append(search)

    return nested_config


def convert_primitive(struct: Primitive, path: Sequence[Index], primitives_config: PrimitivesConfig) -> Primitive:
    for search_path, func in primitives_config:
        if len(search_path) == len(path):
            try:
                struct = func(struct)
            except ValueError as e:
                raise ValueError(f"{str(e)} in section {utils.pretty_path(path)}") from None

    return struct


class NodePrimitives(jinja.JinjaNode):
    """
    A Node that both compiles template strings and converts primitive values in the underlying `dict` or `list`.
    """

    def __init__(self, delegate: Union[dict, list], *,
                 path: jinja.Path, env: jinja.Environment,
                 primitives_config: PrimitivesConfig):
        super().__init__(delegate, path=path, env=env)
        self.primitives_config = primitives_config

    def _child(self, index: Index, val: Union[list, dict]) -> jinja.Node:
        primitives_config = self._nested_primitives_config(index)
        return self._child_type(index)(val, path=self.path + (index,), env=self.env,
                                       primitives_config=primitives_config)

    def _child_type(self, _: Index) -> Type['NodePrimitives']:
        return NodePrimitives

    def _convert(self, index: Index, val: Primitive) -> Primitive:
        val = super()._convert(index, val)

        primitives_config = self._nested_primitives_config(index)
        val = convert_primitive(val, self.path + (index,), primitives_config)

        return val

    def _nested_primitives_config(self, index: Index) -> PrimitivesConfig:
        depth = len(self.path)
        return choose_nested_primitives_config(self.primitives_config, depth, index)


class NodesCustomization:
    """
    Customize access to the particular sections of the inventory.
    """

    # pylint: disable=no-self-argument

    def __init__(nodes) -> None:
        # The classes below should customize access to the sections of the inventory,
        # while preserving the global behaviour of Node implementations: NodePrimitives, JinjaNode, Node.

        class Kubeadm(Node):
            def descend(self, index: Index) -> Union[Primitive, Node]:
                child: Union[Primitive, Node] = super().descend(index)

                if index == 'kubernetesVersion':
                    kubernetes.verify_allowed_version(cast(str, child))

                return child

        class Services(Node):
            def _child_type(self, index: Index) -> Type[Node]:
                if index == 'kubeadm':
                    return nodes.Kubeadm

                return super()._child_type(index)

        class Root(Node):
            def _child_type(self, index: Index) -> Type[Node]:
                if index == 'services':
                    return nodes.Services

                return super()._child_type(index)

        nodes.Kubeadm: Type[Node] = Kubeadm
        nodes.Services: Type[Node] = Services
        nodes.Root: Type[Node] = Root

    def derive(nodes, Base: Type[Node], delegate: dict, **kwargs: Any) -> Node:
        nodes.Kubeadm = cast(Type[Node], type("Kubeadm", (nodes.Kubeadm, Base), {}))
        nodes.Services = cast(Type[Node], type("Services", (nodes.Services, Base), {}))
        nodes.Root = cast(Type[Node], type("Root", (nodes.Root, Base), {}))

        return nodes.Root(delegate, **kwargs)


class Environment(jinja.Environment):
    """
    Environment that supports recursive compilation and on-the-fly conversion of primitive values.

    It also customizes access to the particular sections of the inventory.
    """

    def __init__(self, logger: log.EnhancedLogger, recursive_values: dict,
                 *,
                 recursive_compile: bool = False,
                 recursive_extra: Dict[str, Any] = None,
                 primitives_config: PrimitivesConfig = None):
        """
        Instantiate new environment and set default filters.

        :param logger: EnhancedLogger
        :param recursive_values: The render values access to which should be customized.
                                 They may also be automatically converted and compiled if necessary.
        :param recursive_compile: Flag that enables recursive compilation.
        :param recursive_extra: If recursive compilation occurs, these render values are supplied to the template.
        :param primitives_config: List of sections and convertors of primitive values.
        """
        self.recursive_compile = recursive_compile
        self.primitives_config = primitives_config
        super().__init__(logger, recursive_values, recursive_extra=recursive_extra)

    def create_root(self, delegate: dict) -> Node:
        kwargs = {}
        Base: Type[Node]
        if not self.recursive_compile:
            Base = Node
        else:
            Base = jinja.JinjaNode
            kwargs = {"path": jinja.Path(), "env": self}
            if self.primitives_config is not None:
                Base = NodePrimitives
                kwargs = {**kwargs, "primitives_config": self.primitives_config}

        return NodesCustomization().derive(Base, delegate, **kwargs)



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import os
import shlex
import sys
import time
from abc import abstractmethod, ABC
from copy import deepcopy
from typing import Optional, List, Union, Sequence, Tuple, Dict, Any

from kubemarine.core import utils, cluster as c, action, resources as res, errors, summary, log, defaults

ERROR_UNRECOGNIZED_CUMULATIVE_POINT_EXCLUDE = "Unrecognized cumulative point to exclude: {point}"
ERROR_UNRECOGNIZED_TASKS_FILTER = "Unrecognized tasks filter: {tasks}"

TASK_DESCRIPTION_TEMPLATE = """
tasks list:
    %s
"""

END_OF_TASKS = object()


class FlowResult:
    def __init__(self, context: dict, logger: log.EnhancedLogger):
        self.context = context
        self.logger = logger


class Flow(ABC):
    def run_flow(self, context: Union[dict, res.DynamicResources], print_summary: bool = True) -> FlowResult:
        time_start = time.time()

        if isinstance(context, res.DynamicResources):
            resources = context
        else:
            resources = res.RESOURCES_FACTORY(context)

        context = resources.context

        try:
            utils.prepare_dump_directory(context)
            resources.logger()
            self._run(resources)
        except (Exception, KeyboardInterrupt) as exc:
            logger = resources.logger_if_initialized()
            if isinstance(exc, errors.FailException):
                # pylint: disable-next=no-member
                utils.do_fail(exc.message, exc.reason, exc.hint, logger=logger)
            else:
                utils.do_fail(f"'{context['initial_procedure'] or 'undefined'}' procedure failed.", exc,
                              logger=logger)

        time_end = time.time()
        logger = resources.logger()

        if print_summary:
            summary.schedule_report(resources.result_context, summary.SummaryItem.EXECUTION_TIME,
                                    utils.get_elapsed_string(time_start, time_end))
            summary.print_summary(resources.result_context, logger)
            logger.info("SUCCESSFULLY FINISHED")

        return FlowResult(resources.result_context, logger)

    @abstractmethod
    def _run(self, resources: res.DynamicResources) -> None:
        pass


class ActionsFlow(Flow):
    def __init__(self, actions: List[action.Action]):
        self._actions = actions

    def _run(self, resources: res.DynamicResources) -> None:
        run_actions(resources, self._actions)


def run_actions(resources: res.DynamicResources, actions: Sequence[action.Action]) -> None:
    """
    Runs actions one by one, recreates inventory when necessary,
    managing such resources as cluster object and inventory.

    Preserve inventory each time it is recreated, or in the end if any actions are successful.
    """

    context = resources.context
    logger = resources.logger()

    successfully_performed: List[str] = []
    cluster: Optional[c.KubernetesCluster] = None
    for act in actions:
        act.prepare_context(context)

        if not successfully_performed:
            # first action in group
            if resources.inventory_filepath:
                with utils.open_external(resources.inventory_filepath, "r") as stream:
                    utils.dump_file(context, stream, "cluster_initial.yaml")

            if resources.procedure_inventory_filepath:
                with utils.open_external(resources.procedure_inventory_filepath, "r") as stream:
                    utils.dump_file(context, stream, "procedure.yaml")

        # Initialize connections early, in particular for inventory preservation.
        resources.cluster(c.EnrichmentStage.LIGHT)

        try:
            logger.info(f"Running action '{act.identifier}'")
            act.run(resources)
            resources.collect_action_result()
            successfully_performed.append(act.identifier)
        except (Exception, KeyboardInterrupt):  # even on KeyboardInterrupt we have to preserve what we have done
            if successfully_performed:
                _post_process_actions_group(resources, cluster, successfully_performed, failed=True)

            raise

        cluster = resources.cluster_if_initialized()

        if act.recreate_inventory:
            if resources.inventory_filepath:
                with utils.open_external(resources.inventory_filepath, "r") as stream:
                    # write original file data to backup file with timestamp
                    timestamp = utils.get_current_timestamp_formatted()
                    inventory_file_basename = os.path.basename(resources.inventory_filepath)
                    utils.dump_file(context, stream, "%s_%s" % (inventory_file_basename, str(timestamp)))

            resources.recreate_inventory()
            _post_process_actions_group(resources, cluster, successfully_performed)
            successfully_performed = []
            cluster = None

    if successfully_performed:
        _post_process_actions_group(resources, cluster, successfully_performed)


def _post_process_actions_group(resources: res.DynamicResources, cluster: Optional[c.KubernetesCluster],
                                successfully_performed: List[str],
                                *,
                                failed: bool = False) -> None:
    previous_successful_cluster = cluster is not None
    try:
        if previous_successful_cluster:
            _dump_inventory(resources, failed)
    finally:
        _preserve_inventory(resources, successfully_performed, failed=failed, enriched=previous_successful_cluster)


def _dump_inventory(resources: res.DynamicResources, failed: bool) -> None:
    context = resources.context

    # If cluster is initialized, it is fully enriched at least to DEFAULT state.
    # Use this state to dump all effective inventories.
    # This is acceptable due to the following assumptions for the last action:
    # * If successful, changes in the inventory should be moved to this state in DynamicResources.recreate_inventory().
    # * If failed, switch to this state effectively restores the cluster to the state after previous action succeeded.
    cluster = resources.cluster(c.EnrichmentStage.DEFAULT)

    if failed:
        # Preserve effective inventory for the last succeeded action.
        # For debug aims, cluster_procedure.yaml can still be used.
        defaults.dump_inventory(cluster, context, 'cluster.yaml')

    resources.dump_finalized_inventory(cluster)


def _preserve_inventory(resources: res.DynamicResources, successfully_performed: List[str],
                        *,
                        failed: bool, enriched: bool) -> None:
    context = resources.context

    context['successfully_performed'] = successfully_performed
    context['status'] = 'failed' if failed else 'successful'

    if (resources.context['preserve_inventory']
            and not resources.context['execution_arguments'].get('without_act', False)):
        # Light cluster is always pre-initialized before running of any action.
        cluster = resources.cluster(c.EnrichmentStage.LIGHT)
        cluster.preserve_inventory(context, enriched=enriched)


class TasksAction(action.Action):
    def __init__(self, identifier: str, tasks: dict,
                 *,
                 cumulative_points: dict = None,
                 tasks_filter: List[str] = None,
                 recreate_inventory: bool = False):
        super().__init__(identifier, recreate_inventory=recreate_inventory)
        self.tasks = deepcopy(tasks)
        self.cumulative_points = cumulative_points or {}
        self.tasks_filter = tasks_filter

    def run(self, resources: res.DynamicResources) -> None:
        """
        Filters and runs tasks.
        """

        args: dict = resources.context['execution_arguments']

        check_cumulative_points(args, self.cumulative_points)

        joined_tasks_filter = ','.join(self.tasks_filter) if self.tasks_filter is not None else args.get('tasks')
        excluded_tasks = args.get('exclude')

        logger = resources.logger()
        logger.debug("Excluded tasks:")
        filtered_tasks, final_list = filter_flow(self.tasks, joined_tasks_filter, excluded_tasks, logger)
        if filtered_tasks == self.tasks:
            logger.debug("\tNo excluded tasks")

        cluster = self.cluster(resources)

        if args.get('without_act', False):
            cluster.log.debug('\nFurther acting manually disabled')
            return

        init_tasks_flow(cluster)
        run_tasks_recursive(self.tasks, final_list, cluster, self.cumulative_points, [])
        proceed_cumulative_point(cluster, self.cumulative_points, END_OF_TASKS,
                                 force=args.get('force_cumulative_points', False))

    def cluster(self, _res: res.DynamicResources) -> c.KubernetesCluster:
        return _res.cluster()


def run_tasks(resources: res.DynamicResources, tasks: dict, cumulative_points: dict = None) -> None:
    return TasksAction("", tasks, cumulative_points=cumulative_points).run(resources)


def create_empty_context(args: dict, procedure: str) -> dict:
    return {
        "execution_arguments": deepcopy(args),
        'initial_procedure': procedure,
        'preserve_inventory': True,
        'make_finalized_inventory': True,
        'load_inventory_silent': False,
        'runtime_vars': {},
        'result': ['summary_report'],
    }


def get_task_list(tasks: dict, _task_path: str = '', leafs_only: bool = True) -> List[str]:
    result = []
    for task_name, task in tasks.items():
        __task_path = _task_path + "." + task_name if _task_path != '' else task_name
        if callable(task) or not leafs_only:
            result.append(__task_path)

        if not callable(task):
            result.extend(get_task_list(task, __task_path, leafs_only))

    return result


def create_context(parser: argparse.ArgumentParser, cli_arguments: Optional[list], procedure: str) -> dict:
    args_list = sys.argv[1:]
    if cli_arguments is not None:
        args_list = cli_arguments

    parser.prog = procedure
    args = vars(parse_args(parser, args_list))

    context = create_empty_context(args=args, procedure=procedure)
    context["initial_cli_arguments"] = ' '.join(map(shlex.quote, args_list))

    return context


def split_strings_list(string_list: Optional[str]) -> List[str]:
    if string_list is None:
        string_list = ''

    return [] if not string_list.strip() else \
        list(filter(bool, map(str.strip, string_list.split(","))))


def check_cumulative_points(args: dict, cumulative_points: dict) -> None:
    exclude_points = args.get('exclude_cumulative_points_methods')
    if not isinstance(exclude_points, list):
        exclude_points = split_strings_list(exclude_points)

    points_list = [point_method.__module__ + '.' + point_method.__qualname__ for point_method in cumulative_points]

    for exclude_point in exclude_points:
        if exclude_point not in points_list:
            raise Exception(ERROR_UNRECOGNIZED_CUMULATIVE_POINT_EXCLUDE.format(point=exclude_point))

    args['exclude_cumulative_points_methods'] = exclude_points


def check_tasks_filter(tasks: dict, tasks_filter: List[str]) -> List[str]:
    tasks_list = get_task_list(tasks, leafs_only=False)

    for tasks_ in tasks_filter:
        if tasks_ not in tasks_list:
            raise Exception(ERROR_UNRECOGNIZED_TASKS_FILTER.format(tasks=tasks_))

    return tasks_filter


def filter_flow(tasks: dict, tasks_filter: Optional[str], excluded_tasks: Optional[str],
                logger: log.EnhancedLogger = None) -> Tuple[dict, List[str]]:
    # Remove any whitespaces from filters, and split by '.'
    tasks_path_filter = [tasks_.split(".") for tasks_ in check_tasks_filter(
        tasks, split_strings_list(tasks_filter))]

    excluded_path_tasks = [tasks_.split(".") for tasks_ in check_tasks_filter(
        tasks, split_strings_list(excluded_tasks))]

    return _filter_flow_internal(tasks, tasks_path_filter, excluded_path_tasks, [], logger)


def _filter_flow_internal(tasks: dict, tasks_filter: List[List[str]], excluded_tasks: List[List[str]],
                          _task_path: List[str],
                          logger: log.EnhancedLogger = None) -> Tuple[dict, List[str]]:
    filtered = {}
    final_list = []

    for task_name, task in tasks.items():
        __task_path = _task_path + [task_name]
        __task_name = ".".join(__task_path)

        allowed = True
        # if task_filter is not empty - smb specified filter argument
        if tasks_filter:
            allowed = False
            # Check if the iterable subpath is in allowed paths. For example, we have to check if
            # ['prepare'] is in allowed path ['prepare', 'cri']
            for task_path in tasks_filter:
                # one of task_path, __task_path is a sublist of another
                # check if current '__task_path' is a sublist only if 'task' is not a final task.
                if (task_path[:len(__task_path)] == __task_path and not callable(task)) \
                        or __task_path[:len(task_path)] == task_path:
                    allowed = True

        if allowed and __task_path not in excluded_tasks:
            if callable(task):
                filtered[task_name] = task
                final_list.append(__task_name)
            else:
                filtered_flow, _final_list = _filter_flow_internal(task, tasks_filter, excluded_tasks, __task_path)
                # there is something to execute in subtree
                if filtered_flow:
                    filtered[task_name] = filtered_flow
                    final_list += _final_list
        elif logger:
            logger.debug("\t%s" % __task_name)

    return filtered, final_list


def run_tasks_recursive(tasks: dict, final_task_names: List[str], cluster: c.KubernetesCluster,
                        cumulative_points: dict, _task_path: List[str]) -> None:
    for task_name, task in tasks.items():
        __task_path = _task_path + [task_name]
        __task_name = ".".join(__task_path)
        run = __task_name in final_task_names

        args = cluster.context['execution_arguments']
        # --force-cumulative-points forcibly run the point only if the related task is going to be executed
        force_cumulative_point = run and args.get('force_cumulative_points', False)
        proceed_cumulative_point(cluster, cumulative_points, __task_name, force=force_cumulative_point)

        if callable(task):
            if not run:
                continue
            cluster.log.info("*** TASK %s ***" % __task_name)
            try:
                task(cluster)
                add_task_to_proceeded_list(cluster, __task_name)
            except (Exception, KeyboardInterrupt) as exc:
                raise errors.FailException(
                    "TASK FAILED %s" % __task_name, exc,
                    hint=cluster.globals['error_handling']['failure_message'] % (sys.argv[0], __task_name)
                )
        else:
            run_tasks_recursive(task, final_task_names, cluster, cumulative_points, __task_path)


def new_common_parser(cli_help: str) -> argparse.ArgumentParser:

    parser = argparse.ArgumentParser(description=cli_help,
                                     formatter_class=argparse.RawTextHelpFormatter)

    parser.add_argument('-c', '--config',
                        default='cluster.yaml',
                        help='define main cluster configuration file')

    parser.add_argument('--ansible-inventory-location',
                        default='ansible-inventory.ini',
                        help='auto-generated ansible-compatible inventory file location')

    parser.add_argument('--dump-location',
                        default='.',
                        help='dump directory for intermediate files')

    parser.add_argument('--disable-dump',
                        action='store_true',
                        help='prevent dump directory creation')

    parser.add_argument('--disable-dump-cleanup',
                        action='store_true',
                        help='prevent dump directory cleaning on process launch')

    parser.add_argument('--log',
                        action='append',
                        nargs='*',
                        help='Logging options, can be specified multiple times')

    parser.add_argument('-w', '--workdir',
                        default='',
                        help='Custom path of the workdir')

    return parser


def new_tasks_flow_parser(cli_help: str, tasks: dict = None) -> argparse.ArgumentParser:
    parser = new_common_parser(cli_help)

    parser.add_argument('--without-act',
                        action='store_true',
                        help='prevent tasks to be executed')

    parser.add_argument('--tasks',
                        default='',
                        help='define comma-separated tasks to be executed')

    parser.add_argument('--exclude',
                        default='',
                        help='exclude comma-separated tasks from execution')

    parser.add_argument('--disable-cumulative-points',
                        action='store_true',
                        help='disable cumulative points execution (use only when you understand what you are doing!)')

    parser.add_argument('--force-cumulative-points',
                        action='store_true',
                        help='force cumulative points execution - they will be executed regardless of whether it was '
                             'scheduled or not (use only when you understand what you are doing!)')

    parser.add_argument('--exclude-cumulative-points-methods',
                        default='',
                        help='comma-separated cumulative points methods names to be excluded from execution')

    # Add tasks list to help section
    if tasks is not None:
        parser.epilog = TASK_DESCRIPTION_TEMPLATE % ('\n    '.join(get_task_list(tasks)))

    return parser


def new_procedure_parser(cli_help: str, optional_config: bool = False, tasks: dict = None) -> argparse.ArgumentParser:
    parser = new_tasks_flow_parser(cli_help, tasks)

    help_msg = 'config file for the procedure'
    if optional_config:
        parser.add_argument('procedure_config', metavar='procedure_config', type=str, help=help_msg, nargs='?')
    else:
        parser.add_argument('procedure_config', metavar='procedure_config', type=str, help=help_msg)

    return parser


def parse_args(parser: argparse.ArgumentParser, arguments: list) -> argparse.Namespace:
    args = parser.parse_args(arguments)

    if args.workdir != '':
        os.chdir(args.workdir)

    return args


def proceed_cumulative_point(cluster: c.KubernetesCluster, points_list: dict,
                             point_task_name: Union[str, object], force: bool = False) -> Dict[str, Any]:
    _check_within_flow(cluster)

    if cluster.context['execution_arguments'].get('disable_cumulative_points', False):
        return {}

    scheduled_methods = cluster.context.get('scheduled_cumulative_points', [])

    results = {}
    for point_method, points_tasks_names in points_list.items():
        if point_task_name in points_tasks_names:

            point_method_fullname = point_method.__module__ + '.' + point_method.__qualname__
            if force:
                cluster.log.verbose('Method %s will be forcibly executed' % point_method_fullname)
            elif point_method not in scheduled_methods:
                cluster.log.verbose('Method %s not scheduled - cumulative point call skipped' % point_method_fullname)
                continue

            cluster.log.info("*** CUMULATIVE POINT %s ***" % point_method_fullname)

            call_result = point_method(cluster)
            if point_method in scheduled_methods:
                scheduled_methods.remove(point_method)
            results[point_method_fullname] = call_result

    return results


def init_tasks_flow(cluster: c.KubernetesCluster) -> None:
    if 'proceeded_tasks' not in cluster.context:
        cluster.context['proceeded_tasks'] = []


def add_task_to_proceeded_list(cluster: c.KubernetesCluster, task_path: str) -> None:
    if not cluster.is_task_completed(task_path):
        cluster.context['proceeded_tasks'].append(task_path)
        utils.dump_file(cluster, "\n".join(cluster.context['proceeded_tasks'])+"\n", 'finished_tasks')


def _check_within_flow(cluster: c.KubernetesCluster, check: bool = True) -> None:
    if check != ('proceeded_tasks' in cluster.context):
        raise NotImplementedError(f"The method is called {'not ' if check else ''}within tasks flow execution")



#!/usr/bin/env python3
# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from collections import OrderedDict
from typing import List

from kubemarine import admission
from kubemarine.core import flow

tasks = OrderedDict({
    "manage_pss": admission.manage_pss,
    "restart_pods": admission.restart_pods_task
})


class PSSAction(flow.TasksAction):
    def __init__(self) -> None:
        super().__init__('manage pss', tasks, recreate_inventory=True)


def create_context(cli_arguments: List[str] = None) -> dict:

    cli_help = '''
    Script for managing pss on existing Kubernetes cluster.

    How to use:

    '''

    parser = flow.new_procedure_parser(cli_help, tasks=tasks)
    context = flow.create_context(parser, cli_arguments, procedure='manage_pss')
    return context


def main(cli_arguments: List[str] = None) -> None:
    context = create_context(cli_arguments)
    flow.ActionsFlow([PSSAction()]).run_flow(context)


if __name__ == '__main__':
    main()
